# sage_setup: distribution = sagemath-pari
# sage.doctest: needs sage.libs.ntl
"""
Frobenius endomorphisms on `p`-adic fields
"""
# ****************************************************************************
#       Copyright (C) 2013 Xavier Caruso <xavier.caruso@normalesup.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.rings.integer cimport Integer
from sage.rings.infinity import Infinity

from sage.categories.homset import Hom
from sage.structure.element cimport Element
from sage.structure.richcmp cimport (richcmp, rich_to_bool,
                                     richcmp_not_equal)

from sage.rings.morphism cimport RingHomomorphism
from sage.rings.padics.padic_generic import pAdicGeneric


cdef class FrobeniusEndomorphism_padics(RingHomomorphism):
    """
    A class implementing Frobenius endomorphisms on `p`-adic fields.
    """
    def __init__(self, domain, n=1):
        """
        INPUT:

        - ``domain`` -- an unramified `p`-adic field

        - ``n`` -- integer (default: 1)

        .. NOTE::

            `n` may be negative.

        OUTPUT:

        The `n`-th power of the absolute (arithmetic) Frobenius
        endomorphism on ``domain``

        TESTS::

            sage: from sage.rings.padics.morphism import FrobeniusEndomorphism_padics
            sage: K.<a> = Qq(5^3)
            sage: FrobeniusEndomorphism_padics(K)
            Frobenius endomorphism on 5-adic Unramified Extension ... lifting a |--> a^5 on the residue field
            sage: FrobeniusEndomorphism_padics(K,2)
            Frobenius endomorphism on 5-adic Unramified Extension ... lifting a |--> a^(5^2) on the residue field

            sage: FrobeniusEndomorphism_padics(K,a)
            Traceback (most recent call last):
            ...
            TypeError: n (=a + O(5^20)) is not an integer

            sage: K = Qp(5)
            sage: x = polygen(ZZ, 'x')
            sage: L.<pi> = K.extension(x^2 - 5)
            sage: FrobeniusEndomorphism_padics(L)
            Traceback (most recent call last):
            ...
            TypeError: The domain must be unramified

            sage: FrobeniusEndomorphism_padics(QQ)
            Traceback (most recent call last):
            ...
            TypeError: The domain must be an instance of pAdicGeneric
        """
        if not isinstance(domain, pAdicGeneric):
            raise TypeError("The domain must be an instance of pAdicGeneric")
        if domain.absolute_e() != 1:
            raise TypeError("The domain must be unramified")
        try:
            n = Integer(n)
        except (ValueError, TypeError):
            raise TypeError("n (=%s) is not an integer" % n)

        self._degree = domain.absolute_f()
        self._power = n % self._degree
        self._order = self._degree / domain.absolute_f().gcd(self._power)
        RingHomomorphism.__init__(self, Hom(domain, domain))

    cdef dict _extra_slots(self):
        """
        Helper for copying and pickling.

        TESTS::

            sage: R.<x> = QQ[]
            sage: U.<a> = Qp(2).extension(x^2 + x + 1)
            sage: F = U.frobenius_endomorphism(); F
            Frobenius endomorphism on 2-adic Unramified Extension Field in a defined by x^2 + x + 1 lifting a |--> a^2 on the residue field
            sage: copy(F)
            Frobenius endomorphism on 2-adic Unramified Extension Field in a defined by x^2 + x + 1 lifting a |--> a^2 on the residue field
        """
        slots = RingHomomorphism._extra_slots(self)
        slots['_degree'] = self._degree
        slots['_power'] = self._power
        slots['_order'] = self._order
        return slots

    cdef _update_slots(self, dict slots):
        """
        Helper for copying and pickling.

        TESTS::

            sage: R.<x> = ZZ[]
            sage: U.<a> = Zp(2).extension(x^2 + x + 1)
            sage: F = U.frobenius_endomorphism(); F
            Frobenius endomorphism on 2-adic Unramified Extension Ring in a defined by x^2 + x + 1 lifting a |--> a^2 on the residue field
            sage: loads(dumps(F)) == F
            True
        """
        self._degree = slots['_degree']
        self._power = slots['_power']
        self._order = slots['_order']
        RingHomomorphism._update_slots(self, slots)

    def _repr_(self):
        """
        Return a string representation of this endomorphism.

        EXAMPLES::

            sage: K.<a> = Qq(5^3)
            sage: Frob = K.frobenius_endomorphism(); Frob
            Frobenius endomorphism on 5-adic Unramified Extension ... lifting a |--> a^5 on the residue field

            sage: Frob._repr_()
            'Frobenius endomorphism on 5-adic Unramified Extension ... lifting a |--> a^5 on the residue field'
        """
        name = self.domain().variable_name()
        if self._power == 0:
            s = "Identity endomorphism of %s" % self.domain()
        elif self._power == 1:
            s = "Frobenius endomorphism on %s lifting %s |--> %s^%s on the residue field" % (self.domain(), name, name, self.domain().residue_characteristic())
        else:
            s = "Frobenius endomorphism on %s lifting %s |--> %s^(%s^%s) on the residue field" % (self.domain(), name, name, self.domain().residue_characteristic(), self._power)
        return s

    def _repr_short(self):
        """
        Return a short string representation of this endomorphism.

        EXAMPLES::

            sage: K.<a> = Qq(5^3)
            sage: Frob = K.frobenius_endomorphism(); Frob
            Frobenius endomorphism on 5-adic Unramified Extension ... lifting a |--> a^5 on the residue field

            sage: Frob._repr_short()
            'Frob'
        """
        if self._power == 0:
            s = "Identity"
        elif self._power == 1:
            s = "Frob"
        else:
            s = "Frob^%s" % self._power
        return s

    cpdef Element _call_(self, x):
        """
        TESTS::

            sage: K.<a> = Qq(5^3)
            sage: Frob = K.frobenius_endomorphism()
            sage: Frob(a) == a.frobenius()
            True
        """
        res = x
        for i in range(self._power):
            res = res.frobenius()
        return res

    def order(self):
        """
        Return the order of this endomorphism.

        EXAMPLES::

            sage: K.<a> = Qq(5^12)
            sage: Frob = K.frobenius_endomorphism()
            sage: Frob.order()
            12
            sage: (Frob^2).order()
            6
            sage: (Frob^9).order()
            4
        """
        if self._order == 0:
            return Infinity
        else:
            return Integer(self._order)

    def power(self):
        """
        Return the smallest integer `n` such that this endomorphism
        is the `n`-th power of the absolute (arithmetic) Frobenius.

        EXAMPLES::

            sage: K.<a> = Qq(5^12)
            sage: Frob = K.frobenius_endomorphism()
            sage: Frob.power()
            1
            sage: (Frob^9).power()
            9
            sage: (Frob^13).power()
            1
        """
        return self._power

    def __pow__(self, n, modulus):
        """
        Return the `n`-th iterate of this endomorphism.

        EXAMPLES::

            sage: K.<a> = Qq(5^12)
            sage: Frob = K.frobenius_endomorphism()
            sage: Frob^2
            Frobenius endomorphism on 5-adic Unramified Extension ... lifting a |--> a^(5^2) on the residue field

        The result is simplified if possible::

            sage: Frob^15
            Frobenius endomorphism on 5-adic Unramified Extension ... lifting a |--> a^(5^3) on the residue field
            sage: Frob^36
            Identity endomorphism of 5-adic Unramified Extension ...
        """
        return self.__class__(self.domain(), self.power()*n)

    def _composition(self, right):
        """
        Return ``self`` o ``right``.

        EXAMPLES::

            sage: K.<a> = Qq(5^12)
            sage: f = K.frobenius_endomorphism(); f
            Frobenius endomorphism on 5-adic Unramified Extension ... lifting a |--> a^5 on the residue field
            sage: g = K.frobenius_endomorphism(2); g
            Frobenius endomorphism on 5-adic Unramified Extension ... lifting a |--> a^(5^2) on the residue field
            sage: f * g
            Frobenius endomorphism on 5-adic Unramified Extension ... lifting a |--> a^(5^3) on the residue field

        The result is simplified if possible::

            sage: f = K.frobenius_endomorphism(9)
            sage: g = K.frobenius_endomorphism(10)
            sage: f * g
            Frobenius endomorphism on 5-adic Unramified Extension ... lifting a |--> a^(5^7) on the residue field
        """
        if isinstance(right, FrobeniusEndomorphism_padics):
            return self.__class__(self.domain(), self._power + right.power())
        return RingHomomorphism._composition(self, right)

    def is_injective(self) -> bool:
        """
        Return ``True`` since any power of the Frobenius endomorphism
        over an unramified `p`-adic field is always injective.

        EXAMPLES::

            sage: K.<a> = Qq(5^3)
            sage: Frob = K.frobenius_endomorphism()
            sage: Frob.is_injective()
            True
        """
        return True

    def is_surjective(self) -> bool:
        """
        Return ``True`` since any power of the Frobenius endomorphism
        over an unramified `p`-adic field is always surjective.

        EXAMPLES::

            sage: K.<a> = Qq(5^3)
            sage: Frob = K.frobenius_endomorphism()
            sage: Frob.is_surjective()
            True
        """
        return True

    def is_identity(self) -> bool:
        """
        Return ``True`` if this morphism is the identity morphism.

        EXAMPLES::

            sage: K.<a> = Qq(5^3)
            sage: Frob = K.frobenius_endomorphism()
            sage: Frob.is_identity()
            False
            sage: (Frob^3).is_identity()
            True
        """
        return self.power() == 0

    def __hash__(self):
        r"""
        Return a hash of this morphism.

        It is the hash of ``(domain, codomain, ('Frob', power)``
        where ``power`` is the smallest integer `n` such that
        this morphism acts by `x \mapsto x^(p^n)` on the residue field.

        EXAMPLES::

            sage: K.<a> = Qq(5^3)
            sage: Frob = K.frobenius_endomorphism()
            sage: hash(Frob)  # indirect doctest, random
            2818440606874670810
        """
        domain = self.domain()
        codomain = self.codomain()
        return hash((domain, codomain, ('Frob', self._power)))

    cpdef _richcmp_(left, right, int op):
        """
        Compare ``left`` and ``right``.

        EXAMPLES::

            sage: K.<a> = Qq(5^3)
            sage: F = K.frobenius_endomorphism()
            sage: G = K.frobenius_endomorphism(4)

            sage: F == G
            True
        """
        if left is right:
            return rich_to_bool(op, 0)
        l_domain = left.domain()
        r_domain = right.domain()
        if l_domain != r_domain:
            return richcmp_not_equal(l_domain, r_domain, op)

        l_codomain = left.codomain()
        r_codomain = right.codomain()
        if l_codomain != r_codomain:
            return richcmp_not_equal(l_codomain, r_codomain, op)

        if isinstance(right, FrobeniusEndomorphism_padics):
            return richcmp(left._power, (<FrobeniusEndomorphism_padics>right)._power, op)

        try:
            for x in l_domain.gens():
                lx = left(x)
                rx = right(x)
                if lx != rx:
                    return richcmp_not_equal(lx, rx, op)
            return rich_to_bool(op, 0)
        except (AttributeError, NotImplementedError):
            raise NotImplementedError



# sage_setup: distribution = sagemath-pari
# distutils: libraries = gmp
# sage.doctest: needs sage.rings.real_mpfr
"""
Enumeration of totally real fields: data

AUTHORS:

- John Voight (2007-09-01): Initial version
- John Voight (2007-09-19): various optimization tweaks
- John Voight (2007-10-09): improvements: Smyth bound, Lagrange multipliers for b
- Craig Citro and John Voight (2007-11-04): type checking and other polishing
"""

#*****************************************************************************
#       Copyright (C) 2007 William Stein and John Voight
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from libc.math cimport sqrt
from cysignals.memory cimport sig_malloc, sig_free

from sage.arith.misc import binomial
from sage.arith.misc import GCD as gcd
from sage.libs.gmp.mpz cimport *
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.real_mpfr import RealField
from sage.rings.integer_ring import ZZ
from sage.rings.integer cimport Integer

# Other global variables
ZZx = PolynomialRing(ZZ, 'x')

from libc.math cimport lrint, floor, ceil, fabs, round


#*********************************************************************
# Auxiliary routine
# Hermite constant, naive Newton-Raphson, and a specialized Lagrange
# multiplier solver.
#*********************************************************************

def hermite_constant(n):
    r"""
    Return the `n`-th Hermite constant.

    The `n`-th Hermite constant (typically denoted `\gamma_n`), is defined
    to be

    .. MATH::

        \max_L \min_{0 \neq x \in L} ||x||^2

    where `L` runs over all lattices of dimension `n` and determinant `1`.

    For `n \leq 8` it returns the exact value of `\gamma_n`, and for
    `n > 9` it returns an upper bound on `\gamma_n`.

    INPUT:

    - ``n`` -- integer

    OUTPUT:

    (an upper bound for) the Hermite constant `\gamma_n`

    EXAMPLES::

        sage: hermite_constant(1) # trivial one-dimensional lattice
        1.0
        sage: hermite_constant(2) # Eisenstein lattice
        1.1547005383792515
        sage: 2/sqrt(3.)
        1.15470053837925
        sage: hermite_constant(8) # E_8
        2.0

    .. NOTE::

        The upper bounds used can be found in [CS1999]_ and [CE2003]_.

    AUTHORS:

    - John Voight (2007-09-03)
    """

    if n <= 8:
        # Exact values are known for gamma_n.
        gamman = [1, 1, 4./3, 2, 4, 8, 64./3, 64, 256][n]
        gamma = gamman**(1./n)
    elif n <= 36:
        gamma = [2.13263235569928, 2.26363016185702, 2.39334691240146,
                 2.52178702088414, 2.64929462619823, 2.77580405570023,
                 2.90147761892077, 3.02639364467182, 3.15067928476872,
                 3.27433066745617, 3.39744386110070, 3.52006195697466,
                 3.64224310140724, 3.76403701226104, 3.88547626036618,
                 4.00659977840648, 4.12744375027069, 4.24804458298177,
                 4.36843113799634, 4.48863097933934, 4.60866759008263,
                 4.72856660611662, 4.84834821242630, 4.96803435811402,
                 5.08764086822471, 5.20718687262715, 5.32668836123079,
                 5.44615801810606][n-9]
    else:
        # Mordell's inequality.
        gamma = 5.44615801810606**((n-1.)/35)

    return gamma


cdef double eval_seq_as_poly(int *f, int n, double x) noexcept:
    r"""
    Evaluates the sequence a, thought of as a polynomial with

    .. MATH::

        f[n]*x^n + f[n-1]*x^(n-1) + ... + f[0].
    """
    cdef double s

    # Horner's method: With polynomials of small degree, we shouldn't
    # expect asymptotic methods to be any faster.
    s = f[n]
    for i from n > i >= 0:
        s = s * x + f[i]
    return s

cdef double newton(int *f, int *df, int n, double x0, double eps) noexcept:
    r"""
    Find the real root x of f (with derivative df) near x0
    with provable precision eps, i.e. |x-z| < eps where z is the actual
    root.
    The sequence a corresponds to the polynomial f with

    .. MATH::

        f(x) = x^n + a[n-1]*x^(n-1) + ... + a[0].

    This function (for speed reasons) has no error checking and no
    guarantees are made as to the convergence; a naive Newton-Raphson
    method is used.
    """
    cdef double x, rdx, dx, fx

    x = x0
    dx = eval_seq_as_poly(f,n,x)/eval_seq_as_poly(df,n-1,x)
    x -= dx
    while fabs(dx) > eps:
        # In truly optimized code, one could tune by automatically
        # iterating a certain number of times based on the size of dx to
        # save on a few comparisons.
        # This savings should be almost negligible...?
        dx = eval_seq_as_poly(f,n,x)/eval_seq_as_poly(df,n-1,x)
        x -= dx

    # Small hack for improved performance elsewhere: if it is close to an
    # integer, give it full precision as an integer.
    rdx = round(x)
    if fabs(rdx-x) < eps:
        x = rdx

    # Now ensure that either f(x-eps) or f(x+eps) has opposite sign
    # as f(x), which implies that |x-z| < eps.
    fx = eval_seq_as_poly(f,n,x)
    while not (fx == 0 or fx*eval_seq_as_poly(f,n,x+eps) < 0 or
                          fx*eval_seq_as_poly(f,n,x-eps) < 0):
        dx = eval_seq_as_poly(f,n,x)/eval_seq_as_poly(df,n-1,x)
        x -= dx
        fx = eval_seq_as_poly(f,n,x)
    return x

cdef void newton_in_intervals(int *f, int *df, int n, double *beta,
                              double eps, double *rts) noexcept:
    r"""
    Find the real roots of f in the intervals specified by beta:

        (beta[0],beta[1]), (beta[1],beta[2]), ..., (beta[n-1], beta[n])

    Calls newton_kernel, so same provisos apply---in particular,
    each interval should contain a unique simple root.
    Note the derivative *df is passed but is recomputed--this is
    just a way to save a malloc and free for each call.
    """
    cdef int i

    for i from 0 <= i < n:
        df[i] = f[i+1]*(i+1)
    for i from 0 <= i < n:
        rts[i] = newton(f, df, n, (beta[i]+beta[i+1])/2, eps)

cpdef lagrange_degree_3(int n, int an1, int an2, int an3):
    r"""
    Private function.  Solves the equations which arise in the Lagrange multiplier
    for degree 3: for each `1 \leq r \leq n-2`, we solve

    .. MATH::

        r*x^i + (n-1-r)\cdot y^i + z^i = s_i \quad (i = 1,2,3)

    where the `s_i` are the power sums determined by the coefficients `a`.
    We output the largest value of `z` which occurs.
    We use a precomputed elimination ideal.

    EXAMPLES::

        sage: from sage.rings.number_field.totallyreal_data import lagrange_degree_3
        sage: ls = lagrange_degree_3(3,0,1,2)
        sage: [RealField(10)(x) for x in ls]
        [-1.0, -1.0]
        sage: lagrange_degree_3(3,6,1,2) # random
        [-5.8878, -5.8878]

    TESTS:

    Check that :issue:`13101` is solved::

        sage: lagrange_degree_3(4,12,19,42)
        [0.0, -1.0]
    """
    cdef long coeffs[7]
    cdef int r, rsq, rcu
    cdef int nr, nrsq, nrcu
    cdef int s1, s1sq, s1cu, s1fo, s2, s2sq, s2cu, s3, s3sq

    RRx = PolynomialRing(RealField(20),'x')

    # Newton's relations.
    s1 = -an1
    s2 = -(an1*s1 + 2*an2)
    s3 = -(an1*s2 + an2*s1 + 3*an3)

    s1sq = s1*s1
    s1cu = s1*s1sq
    s1fo = s1*s1cu
    s2sq = s2*s2
    s2cu = s2*s2sq
    s3sq = s3*s3

    z4minmax = []

    for r from 1 <= r <= n-2:
        nr = n-1-r
        # common subexpressions
        rsq = r*r
        rcu = r*rsq
        nrsq = nr*nr
        nrcu = nr*nrsq

        ## x^6
        coeffs[6] = rcu*nr + rcu + 2*rsq*nrsq + 5*rsq*nr + 3*rsq + \
                    r*nrcu + 5*r*nrsq + 7*r*nr + 3*r + nrcu + \
                    3*nrsq + 3*nr + 1

        ## x^5
        coeffs[5] = -6*rsq*nr*s1 - 6*rsq*s1 - 6*r*nrsq*s1 - 18*r*nr*s1 - \
                    12*r*s1 - 6*nrsq*s1 - 12*nr*s1 - 6*s1

        ## x^4
        coeffs[4] = -3*rcu*s2 - 3*rsq*nr*s2 + 3*rsq*s1sq - 6*rsq*s2 - \
                    3*r*nrsq*s2 + 15*r*nr*s1sq - 6*r*nr*s2 + 18*r*s1sq - \
                    3*r*s2 - 3*nrcu*s2 + 3*nrsq*s1sq - 6*nrsq*s2 + \
                    18*nr*s1sq - 3*nr*s2 + 15*s1sq

        ## x^3
        coeffs[3] = -2*rcu*nr*s3 - 4*rsq*nrsq*s3 + 6*rsq*nr*s1*s2 - \
                    6*rsq*nr*s3 + 12*rsq*s1*s2 - 2*r*nrcu*s3 + \
                    6*r*nrsq*s1*s2 - 6*r*nrsq*s3 - 4*r*nr*s1cu + \
                    12*r*nr*s1*s2 - 4*r*nr*s3 - 12*r*s1cu + 12*r*s1*s2 + \
                    12*nrsq*s1*s2 - 12*nr*s1cu + 12*nr*s1*s2 - \
                    20*s1cu

        ## x^2
        coeffs[2] = 3*rcu*s2sq + 6*rsq*nr*s1*s3 - 3*rsq*nr*s2sq - \
                    6*rsq*s1sq*s2 + 3*rsq*s2sq + 6*r*nrsq*s1*s3 - \
                    3*r*nrsq*s2sq - 6*r*nr*s1sq*s2 + 12*r*nr*s1*s3 + \
                    3*r*nr*s2sq + 3*r*s1fo - 18*r*s1sq*s2 + \
                    3*nrcu*s2sq - 6*nrsq*s1sq*s2 + 3*nrsq*s2sq + \
                    3*nr*s1fo - 18*nr*s1sq*s2 + 15*s1fo

        ## x^1
        coeffs[1] = 6*rsq*nr*s2*s3 - 6*rsq*s1*s2sq + 6*r*nrsq*s2*s3 - \
                    12*r*nr*s1sq*s3 - 6*r*nr*s1*s2sq + 12*r*s1cu*s2 - \
                    6*nrsq*s1*s2sq + 12*nr*s1cu*s2 - 6*s1*s1fo

        ## x^0
        coeffs[0] = rcu*nr*s3sq - rcu*s2cu + 2*rsq*nrsq*s3sq - \
                    6*rsq*nr*s1*s2*s3 + rsq*nr*s2cu + 3*rsq*s1sq*s2sq + \
                    r*nrcu*s3sq - 6*r*nrsq*s1*s2*s3 + r*nrsq*s2cu + \
                    4*r*nr*s1cu*s3 + 3*r*nr*s1sq*s2sq - \
                    3*r*s1fo*s2 - nrcu*s2cu + \
                    3*nrsq*s1sq*s2sq - 3*nr*s1fo*s2 + \
                    s1sq*s1fo

        fcoeff = [ int(coeffs[i]) for i in range(7) ]
        f = ZZx(fcoeff)
        df = ZZx([i*coeffs[i] for i in range(1,7)])
        f = f//gcd(f, df)
        fcoeff = [int(c) for c in f.list()]

        rts = RRx(fcoeff).roots()

        if rts:
            rts = [rtsi[0] for rtsi in rts]
            z4minmax = [min(rts + z4minmax), max(rts + z4minmax)]

    if not z4minmax:
        return [0.0, -1.0]

    return z4minmax

cdef int __len_primes = 46
cdef long primessq[46]
primessq_py = [4, 9, 25, 49, 121, 169, 289, 361, 529, 841, 961, 1369, 1681, 1849, 2209, 2809, 3481, 3721, 4489, 5041, 5329, 6241, 6889, 7921, 9409, 10201, 10609, 11449, 11881, 12769, 16129, 17161, 18769, 19321, 22201, 22801, 24649, 26569, 27889, 29929, 32041, 32761, 36481, 37249, 38809, 39601]
for i from 0 <= i < 46:
    primessq[i] = primessq_py[i]


def int_has_small_square_divisor(sage.rings.integer.Integer d):
    r"""
    Return the largest `a` such that `a^2` divides `d` and `a` has prime divisors `< 200`.

    EXAMPLES::

        sage: from sage.rings.number_field.totallyreal_data import int_has_small_square_divisor
        sage: int_has_small_square_divisor(500)
        100
        sage: is_prime(691)
        True
        sage: int_has_small_square_divisor(691)
        1
        sage: int_has_small_square_divisor(691^2)
        1
    """

    cdef int i
    cdef Integer asq

    asq = ZZ(1)
    for i from 0 <= i < __len_primes:
        while mpz_divisible_ui_p(d.value, primessq[i]):
            asq *= primessq[i]
            mpz_divexact_ui(d.value, d.value, primessq[i])

    return asq


cdef int eval_seq_as_poly_int(int *f, int n, int x) noexcept:
    r"""
    Evaluates the sequence a, thought of as a polynomial with

    .. MATH::

        f[n]*x^n + f[n-1]*x^(n-1) + ... + f[0].
    """
    cdef int s

    s = f[n]
    for i from n > i >= 0:
        s = s * x + f[i]
    return s

cdef double eps_abs, phi, sqrt2
eps_abs = 10.**(-12)
phi = 0.618033988749895
sqrt2 = 1.41421356237310

cdef int easy_is_irreducible(int *a, int n) noexcept:
    r"""
    Very often, polynomials have roots in {+/-1, +/-2, +/-phi, sqrt2}, so we rule
    these out quickly.  Returns 0 if reducible, 1 if inconclusive.
    """
    cdef int s, t, st, i

    # Check if a has a root in {1,-1,2,-2}.
    if eval_seq_as_poly_int(a,n,1) == 0 or eval_seq_as_poly_int(a,n,-1) == 0 or eval_seq_as_poly_int(a,n,2) == 0 or eval_seq_as_poly_int(a,n,-2) == 0:
        return 0

    # Check if f has factors x^2-x-1, x^2+x-1, x^2-2, respectively.
    # Note we only call the ZZx constructor if we're almost certain to reject.
    if fabs(eval_seq_as_poly(a,n,-phi)) < eps_abs:
        s = 2*a[n]
        t = 0
        for i from n > i >= 0:
            st = (s+t)//2
            s = 2*t+st+2*a[i]
            t = st
        if s == 0 and t == 0:
            return 0
    if fabs(eval_seq_as_poly(a,n,phi)) < eps_abs:
        s = 2*a[n]
        t = 0
        for i from n > i >= 0:
            st = (s-t)//2
            s = 2*t-st+2*a[i]
            t = st
        if s == 0 and t == 0:
            return 0
    if fabs(eval_seq_as_poly(a,n,sqrt2)) < eps_abs:
        s = a[n]
        t = 0
        for i from n > i >= 0:
            st = s
            s = 2*t+a[i]
            t = st
        if s == 0 and t == 0:
            return 0

    return 1


def easy_is_irreducible_py(f):
    """
    Used solely for testing easy_is_irreducible.

    EXAMPLES::

        sage: from sage.rings.number_field.totallyreal_data import easy_is_irreducible_py
        sage: easy_is_irreducible_py(pari('x^2+1'))
        1
        sage: easy_is_irreducible_py(pari('x^2-1'))
        0
    """
    cdef int a[10]

    for i from 0 <= i < len(f):
        a[i] = f[i]
    return easy_is_irreducible(a, len(f)-1)


#****************************************************************************
# Main class and routine
#****************************************************************************

# Global precision to find roots; this should probably depend on the
# architecture in some way.  Algorithm gives provably correct results
# for any eps, but an optimal value of eps will be neither too large
# (which gives trivial bounds on coefficients) nor too small (which
# spends needless time computing higher precision on the roots).
cdef double eps_global
eps_global = 10.**(-4)

from sage.rings.number_field.totallyreal_phc import __lagrange_bounds_phc

cdef class tr_data:
    r"""
    This class encodes the data used in the enumeration of totally real
    fields.

    We do not give a complete description here.  For more information,
    see the attached functions; all of these are used internally by the
    functions in :mod:`.totallyreal`, so see that file for examples and
    further documentation.
    """

    def __init__(self, int n, B, a=[]):
        r"""
        Initialization routine (constructor).

        INPUT:

        - ``n`` -- integer; the degree
        - ``B`` -- integer; the discriminant bound
        - ``a`` -- list (default: ``[]``); the coefficient list to begin with,
          where ``a[len(a)]*x^n + ... + a[0]x^(n-len(a))``

        OUTPUT:

        the data initialized to begin enumeration of totally real fields
        with degree n, discriminant bounded by B, and starting with
        coefficients a.

        EXAMPLES::

            sage: from sage.rings.number_field.totallyreal_data import tr_data
            sage: T = tr_data(2,100)
            sage: T.printa()
            k = 0
            a = [0, -1, 1]
            amax = [0, 0, 1]
            beta =  [...]
            gnk =  [...]
        """

        cdef int i

        # Initialize constants.
        self.n = n
        self.B = B
        self.gamma = hermite_constant(n-1)

        # Declare the coefficients of the polynomials (and max such).
        self.a = <int*>sig_malloc(sizeof(int)*(n+1))
        if self.a == NULL: raise MemoryError
        self.amax = <int*>sig_malloc(sizeof(int)*(n+1))
        if self.amax == NULL: raise MemoryError
        # df is memory set aside for the derivative, as
        # used in Newton iteration above.
        self.df = <int*>sig_malloc(sizeof(int)*(n+1))
        if self.df == NULL: raise MemoryError

        for i from 0 <= i < n+1:
            self.a[i] = 0
            self.amax[i] = 0
            self.df[i] = 0

        # beta is an array of arrays (of length n) which list the
        # roots of the derivatives.
        self.beta = <double*>sig_malloc(sizeof(double)*n*(n+1))
        if self.beta == NULL: raise MemoryError
        # gnk is the collection of (normalized) derivatives.
        self.gnk = <int*>sig_malloc(sizeof(int)*(n+1)*n)
        if self.gnk == NULL: raise MemoryError

        for i from 0 <= i < (n+1)*n:
            self.beta[i] = <double>0
            self.gnk[i] = 0

        # Initialize variables.
        if a == []:
            # No starting input, all polynomials will be found; initialize to zero.
            a = [0]*n + [1]
            for i from 0 <= i < n+1:
                self.a[i] = a[i]
                self.amax[i] = a[i]
            self.a[n-1] = -(n//2)
            self.amax[n-1] = 0
            self.k = n-2
        elif len(a) <= n+1:
            # First few coefficients have been specified.
            # The value of k is the largest index of the coefficients of a which is
            # currently unknown; e.g., if k == -1, then we can iterate
            # over polynomials, and if k == n-1, then we have finished iterating.
            if a[len(a)-1] != 1:
                raise ValueError("a[len(a)-1](=%s) must be 1 so polynomial is monic" % a[len(a)-1])

            k = n-len(a)
            self.k = k
            a = [0]*(k+1) + a
            for i from 0 <= i < n+1:
                self.a[i] = a[i]
                self.amax[i] = a[i]

            # Bounds come from an application of Lagrange multipliers in degrees 2,3.
            self.b_lower = -1./n*(self.a[n-1] + (n-1.)*sqrt((1.*self.a[n-1])**2 - 2.*(1+1./(n-1))*self.a[n-2]))
            self.b_upper = -1./n*(self.a[n-1] - (n-1.)*sqrt((1.*self.a[n-1])**2 - 2.*(1+1./(n-1))*self.a[n-2]))
            if k < n-3:
                bminmax = lagrange_degree_3(n,a[n-1],a[n-2],a[n-3])
                if bminmax:
                    self.b_lower = bminmax[0]
                    self.b_upper = bminmax[1]

            # Annoying, but must reverse coefficients for NumPy.
            gnk = [int(binomial(j,k+2))*a[j] for j in range(k+2,n+1)]
            gnk.reverse()
            import numpy
            rts = numpy.roots(gnk).tolist()
            rts.sort()
            self.beta[(k+1)*(n+1)+0] = self.b_lower
            for i from 0 <= i < n-k-2:
                self.beta[(k+1)*(n+1)+(i+1)] = rts[i]
            self.beta[(k+1)*(n+1)+(n-k-1)] = self.b_upper

            # Now to really initialize gnk.
            gnk = [0] + [binomial(j,k+1)*a[j] for j in range(k + 2, n + 1)]
            for i from 0 <= i < n-k:
                self.gnk[(k+1)*n+i] = gnk[i]
        else:
            # Bad input!
            raise ValueError("a has length %s > n+1" % len(a))

    def __dealloc__(self):
        r"""
        Destructor.
        """
        sig_free(self.df)
        sig_free(self.a)
        sig_free(self.amax)
        sig_free(self.beta)
        sig_free(self.gnk)

    def increment(self, verbose=False, haltk=0, phc=False):
        r"""
        'Increment' the totally real data to the next
        value which satisfies the bounds essentially given by Rolle's
        theorem, and return the next polynomial as a sequence of
        integers.

        The default or usual case just increments the constant
        coefficient; then inductively, if this is outside of the
        bounds we increment the next higher coefficient, and so on.

        If there are no more coefficients to be had, returns the zero
        polynomial.

        INPUT:

        - ``verbose`` -- boolean to print verbosely computational details
        - ``haltk`` -- integer; the level at which to halt the inductive
          coefficient bounds
        - ``phc`` -- boolean, if PHCPACK is available, use it when `k = n-5` to
          compute an improved Lagrange multiplier bound

        OUTPUT: the next polynomial, as a sequence of integers

        EXAMPLES::

            sage: from sage.rings.number_field.totallyreal_data import tr_data
            sage: T = tr_data(2,100)
            sage: T.increment()
            [-24, -1, 1]
            sage: for i in range(19): _ = T.increment()
            sage: T.increment()
            [-3, -1, 1]
            sage: T.increment()
            [-25, 0, 1]
        """
        cdef int *f_out
        cdef int i

        f_out = <int *>sig_malloc(sizeof(int) * (self.n + 1))
        if f_out == NULL:
            raise MemoryError("unable to allocate coefficient list")
        for i from 0 <= i < self.n:
            f_out[i] = 0
        f_out[self.n] = 1

        self.incr(f_out, verbose, haltk, phc)

        g = [0] * (1 + self.n)
        for i from 0 <= i <= self.n:
            g[i] = f_out[i]
        sig_free(f_out)

        return g

    cdef void incr(self, int *f_out, int verbose, int haltk, int phc) noexcept:
        r"""
        This function 'increments' the totally real data to the next
        value which satisfies the bounds essentially given by Rolle's
        theorem, and returns the next polynomial in the sequence
        f_out.

        The default or usual case just increments the constant
        coefficient; then inductively, if this is outside of the
        bounds we increment the next higher coefficient, and so on.

        If there are no more coefficients to be had, returns the zero
        polynomial.

        INPUT:

        - ``f_out`` -- integer sequence, to be written with the coefficients of
          the next polynomial
        - ``verbose`` -- boolean to print verbosely computational details
        - ``haltk`` -- integer; the level at which to halt the inductive
          coefficient bounds
        - ``phc`` -- boolean; if PHCPACK is available, use it when ``k == n-5``
          to compute an improved Lagrange multiplier bound

        OUTPUT:

            None. The return value is stored in the variable f_out.
        """
        cdef int n, np1, k, i, nk, kz
        cdef int *gnkm
        cdef int *gnkm1
        cdef double *betak
        cdef double bl, br, akmin, akmax, tmp_dbl
        cdef bint maxoutflag

        n = self.n
        np1 = n+1
        k = self.k

        # If k == -1, we have a full polynomial, so we add 1 to the constant coefficient.
        if k == -1:
            self.a[0] += 1
            # Can't have constant coefficient zero!
            if self.a[0] == 0:
                self.a[0] += 1
            if self.a[0] <= self.amax[0] and easy_is_irreducible(self.a, n):
                for i from 0 <= i < n:
                    f_out[i] = self.a[i]
                return
            else:
                if verbose:
                    print(" ", end="")
                    for i from 0 <= i < np1:
                        print(self.a[i], end="")
                    print(">", end="")
                    for i from 0 <= i < np1:
                        print(self.amax[i], end="")
                    print("")

                # Already reached maximum, so "carry the 1" to find the next value of k.
                k += 1
                while k <= n-1 and self.a[k] >= self.amax[k]:
                    k += 1
                self.a[k] += 1
                self.gnk[n*k] = 0
                k -= 1
        # If we are working through an initialization routine, treat that.
        elif haltk and k == haltk-1:
            self.a[k] += 1
            if self.a[k] > self.amax[k]:
                k += 1
                while k <= n-1 and self.a[k] >= self.amax[k]:
                    k += 1
                self.a[k] += 1
                self.gnk[n*k] = 0
                k -= 1

        # If in the previous step we finished all possible values of
        # the lastmost coefficient, so we must compute bounds on the next coefficient.
        # Recall k == n-1 implies iteration is complete.
        while k < n-1:
            # maxoutflag flags a required abort along the way
            maxoutflag = 0

            # Recall k == -1 means all coefficients are good to go.
            while k >= 0 and (not haltk or k >= haltk):
                if verbose:
                    print(k, ":", end="")
                    for i from 0 <= i < np1:
                        print(self.a[i], end="")
                    print("")

                if k == n-2:
                    # We only know the value of a[n-1], the trace.  Need to apply
                    # basic bounds from Hunter's theorem and Siegel's theorem, with
                    # improvements due to Smyth to get bounds on a[n-2].
                    bl = 1./2*(1-1./n)*(1.*self.a[n-1])**2 \
                         - 1./2*self.gamma*(1./n*self.B)**(1./(n-1))
                    self.a[k] = lrint(ceil(bl))
                    br = 1./2*(1.*self.a[n-1])**2 - 0.88595*n
                    self.amax[k] = lrint(floor(br))

                    # If maximum is already greater than the minimum, break!
                    if self.a[k] > self.amax[k]:
                        if verbose:
                            print(" ", end="")
                            for i from 0 <= i < np1:
                                print(self.a[i], end="")
                            print(">", end="")
                            for i from 0 <= i < np1:
                                print(self.amax[i], end="")
                            print("")
                        maxoutflag = 1
                        break

                    # Knowing a[n-1] and a[n-2] means we can apply bounds from
                    # the Lagrange multiplier in degree 2, which can be solved
                    # immediately.
                    self.b_lower = -1./n*(self.a[n-1] + (n-1.)*sqrt((1.*self.a[n-1])**2 - 2.*(1+1./(n-1))*self.a[n-2]))
                    self.b_upper = -1./n*(self.a[n-1] - (n-1.)*sqrt((1.*self.a[n-1])**2 - 2.*(1+1./(n-1))*self.a[n-2]))

                    # Initialize the second derivative.
                    self.beta[k*np1+0] = self.b_lower
                    self.beta[k*np1+1] = -self.a[n-1]*1./n
                    self.beta[k*np1+2] = self.b_upper
                    self.gnk[k*n+0] = 0
                    self.gnk[k*n+1] = (n-1)*self.a[n-1]
                    self.gnk[k*n+2] = n*(n-1)/2

                    if verbose:
                        print(" ", '%.2f' % self.beta[k * np1 + 1])
                else:
                    # Compute the roots of the derivative.
                    self.gnk[(k+1)*n+0] += self.a[k+1]
                    newton_in_intervals(&self.gnk[(k+1)*n], self.df, n-(k+1),
                                        &self.beta[(k+1)*np1],
                                        eps_global, &self.beta[k*np1+1])
                    if verbose:
                        print(" ", end="")
                        for i from 0 <= i < n-k-1:
                            print('%.2f' % self.beta[k * np1 + 1 + i], end="")
                        print("")

                    for i from 0 <= i < n-k-1:
                        if fabs(self.beta[k*np1+i]
                                 - self.beta[k*np1+(i+1)]) < 10*eps_global:
                            # This happens reasonably infrequently, so calling
                            # the Python routine should be sufficiently fast...
                            f = ZZx([self.gnk[(k+1)*n+i] for i in range(n-(k+1)+1)])
                            # Could just take self.gnk(k+2)*n+i, but this may not be initialized...
                            df = ZZx([i*self.gnk[(k+1)*n+i] for i in range(1,n-(k+1)+1)])
                            if gcd(f,df) != 1:
                                if verbose:
                                    print("  gnk has multiple factor!")
                                maxoutflag = 1
                                break
                    if maxoutflag:
                        break

                    # Bounds come from an application of Lagrange multipliers in degrees 2,3.
                    if k == n-3:
                        self.b_lower = -1./n*(self.a[n-1] + (n-1.)*sqrt((1.*self.a[n-1])**2 - 2.*(1+1./(n-1))*self.a[n-2]))
                        self.b_upper = -1./n*(self.a[n-1] - (n-1.)*sqrt((1.*self.a[n-1])**2 - 2.*(1+1./(n-1))*self.a[n-2]))
                    elif k == n-4:
                        # New bounds from Lagrange multiplier in degree 3.
                        bminmax = lagrange_degree_3(n,self.a[n-1],self.a[n-2],self.a[n-3])
                        if bminmax:
                            self.b_lower = bminmax[0]
                            self.b_upper = bminmax[1]
                    elif k == n-5 and phc:
                        # New bounds using phc/Lagrange multiplier in degree 4.
                        bminmax = __lagrange_bounds_phc(n, 4, [self.a[i] for i from 0 <= i <= n])
                        if bminmax:
                            self.b_lower = bminmax[0]
                            self.b_upper = bminmax[1]
                        else:
                            maxoutflag = 1
                            break

                    if verbose:
                        print("  [LM bounds:", '%.2f' % self.b_lower,
                              '%.2f' % self.b_upper, end="")
                        tb = sqrt((1.*self.a[n-1])**2 - 2.*self.a[n-2])
                        print("vs. +/-", '%.2f' % tb, ']')

                    self.beta[k*np1+0] = self.b_lower
                    self.beta[k*np1+n-k] = self.b_upper

                    # Compute next g_(n-(k+1)), k times the formal integral of g_(n-k).
                    gnkm = &self.gnk[k*n]
                    gnkm1 = &self.gnk[(k+1)*n]
                    gnkm[0] = 0
                    for i from 1 <= i < n-k+1:
                        gnkm[i] = gnkm[n+i-1]*(k+1)/i
                    nk = n-(k+1)

                    # Compute upper and lower bounds which guarantee one retains
                    # a polynomial with all real roots.
                    betak = &self.beta[k*np1]
                    akmin = -eval_seq_as_poly(gnkm, n-k, betak[nk+1]) \
                            -fabs(eval_seq_as_poly(gnkm1, nk, betak[nk+1]))*eps_global
                    for i from 1 <= i < (nk+1)/2+1:
                        # Use the fact that f(z) <= f(x)+|f'(x)|eps if |x-z| < eps
                        # for sufficiently small eps, f(z) = 0, and f''(z) < 0.
                        tmp_dbl = -eval_seq_as_poly(gnkm, n-k, betak[nk+1-2*i]) \
                                  -fabs(eval_seq_as_poly(gnkm1, nk, betak[nk+1-2*i]))*eps_global
                        if tmp_dbl > akmin:
                            akmin = tmp_dbl

                    akmax = -eval_seq_as_poly(gnkm, n-k, betak[nk]) \
                            +fabs(eval_seq_as_poly(gnkm1, n-(k+1), betak[nk]))*eps_global
                    for i from 1 <= i < nk/2+1:
                        # Similar calculus statement here.
                        tmp_dbl = -eval_seq_as_poly(gnkm, n-k, betak[nk-2*i]) \
                                  +fabs(eval_seq_as_poly(gnkm1, nk, betak[nk-2*i]))*eps_global
                        if tmp_dbl < akmax:
                            akmax = tmp_dbl

                    self.a[k] = lrint(ceil(akmin))
                    self.amax[k] = lrint(floor(akmax))

                    if self.a[n-1] == 0 and (n-k) % 2 == 1:
                        # Can replace alpha by -alpha, so if all
                        # "odd" coefficients are zero, may assume next
                        # "odd" coefficient is positive.
                        kz = n-3
                        while kz > k and self.a[kz] == 0:
                            kz -= 2
                        if kz == k:
                            if self.a[k] < 0:
                                self.a[k] = 0
                    if self.a[k] == 0 and self.a[k+1] == 0:
                        self.a[k] += 1
                    # Can't have constant coefficient zero!
                    if k == 0 and self.a[k] == 0:
                        self.a[k] = 1

                    if self.a[k] > self.amax[k]:
                        if verbose:
                            print(" ", end="")
                            for i from 0 <= i < np1:
                                print(self.a[i], end="")
                            print(">", end="")
                            for i from 0 <= i < np1:
                                print(self.amax[i], end="")
                            print("")
                        maxoutflag = 1
                        break

                self.k -= 1
                k -= 1

            if not maxoutflag and easy_is_irreducible(self.a, n):
                self.k = k
                for i from 0 <= i < n:
                    f_out[i] = self.a[i]
                return
            else:
                k += 1
                while k <= n-1 and self.a[k] >= self.amax[k]:
                    k += 1
                self.a[k] += 1
                self.gnk[n*k] = 0
                k -= 1

        # k == n-1, so iteration is complete; return the zero polynomial (of degree n+1).
        self.k = k
        f_out[n] = 0
        return

    def printa(self):
        """
        Print relevant data for ``self``.

        EXAMPLES::

            sage: from sage.rings.number_field.totallyreal_data import tr_data
            sage: T = tr_data(3,2^10)
            sage: T.printa()
            k = 1
            a = [0, 0, -1, 1]
            amax = [0, 0, 0, 1]
            beta =  [...]
            gnk =  [...]
        """
        print("k =", self.k)
        print("a =", [self.a[i] for i in range(self.n + 1)])
        print("amax =", [self.amax[i] for i in range(self.n + 1)])
        print("beta = ", [self.beta[i] for i in range(self.n * (self.n + 1))])
        print("gnk = ", [self.gnk[i] for i in range(self.n * (self.n + 1))])



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/arb.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void arb_init(arb_t x) noexcept
    void arb_clear(arb_t x) noexcept
    arb_ptr _arb_vec_init(slong n) noexcept
    void _arb_vec_clear(arb_ptr v, slong n) noexcept
    void arb_swap(arb_t x, arb_t y) noexcept
    slong arb_allocated_bytes(const arb_t x) noexcept
    slong _arb_vec_allocated_bytes(arb_srcptr vec, slong len) noexcept
    double _arb_vec_estimate_allocated_bytes(slong len, slong prec) noexcept
    void arb_set(arb_t y, const arb_t x) noexcept
    void arb_set_arf(arb_t y, const arf_t x) noexcept
    void arb_set_si(arb_t y, slong x) noexcept
    void arb_set_ui(arb_t y, ulong x) noexcept
    void arb_set_d(arb_t y, double x) noexcept
    void arb_set_fmpz(arb_t y, const fmpz_t x) noexcept
    void arb_set_fmpz_2exp(arb_t y, const fmpz_t x, const fmpz_t e) noexcept
    void arb_set_round(arb_t y, const arb_t x, slong prec) noexcept
    void arb_set_round_fmpz(arb_t y, const fmpz_t x, slong prec) noexcept
    void arb_set_round_fmpz_2exp(arb_t y, const fmpz_t x, const fmpz_t e, slong prec) noexcept
    void arb_set_fmpq(arb_t y, const fmpq_t x, slong prec) noexcept
    int arb_set_str(arb_t res, const char * inp, slong prec) noexcept
    char * arb_get_str(const arb_t x, slong n, ulong flags) noexcept
    void arb_zero(arb_t x) noexcept
    void arb_one(arb_t f) noexcept
    void arb_pos_inf(arb_t x) noexcept
    void arb_neg_inf(arb_t x) noexcept
    void arb_zero_pm_inf(arb_t x) noexcept
    void arb_indeterminate(arb_t x) noexcept
    void arb_zero_pm_one(arb_t x) noexcept
    void arb_unit_interval(arb_t x) noexcept
    void arb_print(const arb_t x) noexcept
    void arb_fprint(FILE * file, const arb_t x) noexcept
    void arb_printd(const arb_t x, slong digits) noexcept
    void arb_fprintd(FILE * file, const arb_t x, slong digits) noexcept
    void arb_printn(const arb_t x, slong digits, ulong flags) noexcept
    void arb_fprintn(FILE * file, const arb_t x, slong digits, ulong flags) noexcept
    char * arb_dump_str(const arb_t x) noexcept
    int arb_load_str(arb_t x, const char * str) noexcept
    int arb_dump_file(FILE * stream, const arb_t x) noexcept
    int arb_load_file(arb_t x, FILE * stream) noexcept
    void arb_randtest(arb_t x, flint_rand_t state, slong prec, slong mag_bits) noexcept
    void arb_randtest_exact(arb_t x, flint_rand_t state, slong prec, slong mag_bits) noexcept
    void arb_randtest_precise(arb_t x, flint_rand_t state, slong prec, slong mag_bits) noexcept
    void arb_randtest_positive(arb_t x, flint_rand_t state, slong prec, slong mag_bits) noexcept
    void arb_randtest_wide(arb_t x, flint_rand_t state, slong prec, slong mag_bits) noexcept
    void arb_randtest_special(arb_t x, flint_rand_t state, slong prec, slong mag_bits) noexcept
    void arb_get_rand_fmpq(fmpq_t q, flint_rand_t state, const arb_t x, slong bits) noexcept
    void arb_urandom(arb_t x, flint_rand_t state, slong prec) noexcept
    void arb_get_mid_arb(arb_t m, const arb_t x) noexcept
    void arb_get_rad_arb(arb_t r, const arb_t x) noexcept
    void arb_add_error_arf(arb_t x, const arf_t err) noexcept
    void arb_add_error_mag(arb_t x, const mag_t err) noexcept
    void arb_add_error(arb_t x, const arb_t err) noexcept
    void arb_add_error_2exp_si(arb_t x, slong e) noexcept
    void arb_add_error_2exp_fmpz(arb_t x, const fmpz_t e) noexcept
    void arb_union(arb_t z, const arb_t x, const arb_t y, slong prec) noexcept
    int arb_intersection(arb_t z, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_nonnegative_part(arb_t res, const arb_t x) noexcept
    void arb_get_abs_ubound_arf(arf_t u, const arb_t x, slong prec) noexcept
    void arb_get_abs_lbound_arf(arf_t u, const arb_t x, slong prec) noexcept
    void arb_get_ubound_arf(arf_t u, const arb_t x, slong prec) noexcept
    void arb_get_lbound_arf(arf_t u, const arb_t x, slong prec) noexcept
    void arb_get_mag(mag_t z, const arb_t x) noexcept
    void arb_get_mag_lower(mag_t z, const arb_t x) noexcept
    void arb_get_mag_lower_nonnegative(mag_t z, const arb_t x) noexcept
    void arb_get_interval_fmpz_2exp(fmpz_t a, fmpz_t b, fmpz_t exp, const arb_t x) noexcept
    void arb_set_interval_mag(arb_t x, const mag_t a, const mag_t b, slong prec) noexcept
    void arb_set_interval_arf(arb_t x, const arf_t a, const arf_t b, slong prec) noexcept
    void arb_set_interval_mpfr(arb_t x, const mpfr_t a, const mpfr_t b, slong prec) noexcept
    void arb_set_interval_neg_pos_mag(arb_t x, const mag_t a, const mag_t b, slong prec) noexcept
    void arb_get_interval_arf(arf_t a, arf_t b, const arb_t x, slong prec) noexcept
    void arb_get_interval_mpfr(mpfr_t a, mpfr_t b, const arb_t x) noexcept
    slong arb_rel_error_bits(const arb_t x) noexcept
    slong arb_rel_accuracy_bits(const arb_t x) noexcept
    slong arb_rel_one_accuracy_bits(const arb_t x) noexcept
    slong arb_bits(const arb_t x) noexcept
    void arb_trim(arb_t y, const arb_t x) noexcept
    int arb_get_unique_fmpz(fmpz_t z, const arb_t x) noexcept
    void arb_floor(arb_t y, const arb_t x, slong prec) noexcept
    void arb_ceil(arb_t y, const arb_t x, slong prec) noexcept
    void arb_trunc(arb_t y, const arb_t x, slong prec) noexcept
    void arb_nint(arb_t y, const arb_t x, slong prec) noexcept
    void arb_get_fmpz_mid_rad_10exp(fmpz_t mid, fmpz_t rad, fmpz_t exp, const arb_t x, slong n) noexcept
    int arb_can_round_arf(const arb_t x, slong prec, arf_rnd_t rnd) noexcept
    int arb_can_round_mpfr(const arb_t x, slong prec, mpfr_rnd_t rnd) noexcept
    bint arb_is_zero(const arb_t x) noexcept
    bint arb_is_nonzero(const arb_t x) noexcept
    bint arb_is_one(const arb_t f) noexcept
    bint arb_is_finite(const arb_t x) noexcept
    bint arb_is_exact(const arb_t x) noexcept
    bint arb_is_int(const arb_t x) noexcept
    bint arb_is_int_2exp_si(const arb_t x, slong e) noexcept
    bint arb_equal(const arb_t x, const arb_t y) noexcept
    bint arb_equal_si(const arb_t x, slong y) noexcept
    bint arb_is_positive(const arb_t x) noexcept
    bint arb_is_nonnegative(const arb_t x) noexcept
    bint arb_is_negative(const arb_t x) noexcept
    bint arb_is_nonpositive(const arb_t x) noexcept
    bint arb_overlaps(const arb_t x, const arb_t y) noexcept
    bint arb_contains_arf(const arb_t x, const arf_t y) noexcept
    bint arb_contains_fmpq(const arb_t x, const fmpq_t y) noexcept
    bint arb_contains_fmpz(const arb_t x, const fmpz_t y) noexcept
    bint arb_contains_si(const arb_t x, slong y) noexcept
    bint arb_contains_mpfr(const arb_t x, const mpfr_t y) noexcept
    bint arb_contains(const arb_t x, const arb_t y) noexcept
    bint arb_contains_int(const arb_t x) noexcept
    bint arb_contains_zero(const arb_t x) noexcept
    bint arb_contains_negative(const arb_t x) noexcept
    bint arb_contains_nonpositive(const arb_t x) noexcept
    bint arb_contains_positive(const arb_t x) noexcept
    bint arb_contains_nonnegative(const arb_t x) noexcept
    bint arb_contains_interior(const arb_t x, const arb_t y) noexcept
    bint arb_eq(const arb_t x, const arb_t y) noexcept
    bint arb_ne(const arb_t x, const arb_t y) noexcept
    bint arb_lt(const arb_t x, const arb_t y) noexcept
    bint arb_le(const arb_t x, const arb_t y) noexcept
    bint arb_gt(const arb_t x, const arb_t y) noexcept
    bint arb_ge(const arb_t x, const arb_t y) noexcept
    void arb_neg(arb_t y, const arb_t x) noexcept
    void arb_neg_round(arb_t y, const arb_t x, slong prec) noexcept
    void arb_abs(arb_t y, const arb_t x) noexcept
    void arb_nonnegative_abs(arb_t y, const arb_t x) noexcept
    void arb_sgn(arb_t y, const arb_t x) noexcept
    int arb_sgn_nonzero(const arb_t x) noexcept
    void arb_min(arb_t z, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_max(arb_t z, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_minmax(arb_t z1, arb_t z2, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_add(arb_t z, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_add_arf(arb_t z, const arb_t x, const arf_t y, slong prec) noexcept
    void arb_add_ui(arb_t z, const arb_t x, ulong y, slong prec) noexcept
    void arb_add_si(arb_t z, const arb_t x, slong y, slong prec) noexcept
    void arb_add_fmpz(arb_t z, const arb_t x, const fmpz_t y, slong prec) noexcept
    void arb_add_fmpz_2exp(arb_t z, const arb_t x, const fmpz_t m, const fmpz_t e, slong prec) noexcept
    void arb_sub(arb_t z, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_sub_arf(arb_t z, const arb_t x, const arf_t y, slong prec) noexcept
    void arb_sub_ui(arb_t z, const arb_t x, ulong y, slong prec) noexcept
    void arb_sub_si(arb_t z, const arb_t x, slong y, slong prec) noexcept
    void arb_sub_fmpz(arb_t z, const arb_t x, const fmpz_t y, slong prec) noexcept
    void arb_mul(arb_t z, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_mul_arf(arb_t z, const arb_t x, const arf_t y, slong prec) noexcept
    void arb_mul_si(arb_t z, const arb_t x, slong y, slong prec) noexcept
    void arb_mul_ui(arb_t z, const arb_t x, ulong y, slong prec) noexcept
    void arb_mul_fmpz(arb_t z, const arb_t x, const fmpz_t y, slong prec) noexcept
    void arb_mul_2exp_si(arb_t y, const arb_t x, slong e) noexcept
    void arb_mul_2exp_fmpz(arb_t y, const arb_t x, const fmpz_t e) noexcept
    void arb_addmul(arb_t z, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_addmul_arf(arb_t z, const arb_t x, const arf_t y, slong prec) noexcept
    void arb_addmul_si(arb_t z, const arb_t x, slong y, slong prec) noexcept
    void arb_addmul_ui(arb_t z, const arb_t x, ulong y, slong prec) noexcept
    void arb_addmul_fmpz(arb_t z, const arb_t x, const fmpz_t y, slong prec) noexcept
    void arb_submul(arb_t z, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_submul_arf(arb_t z, const arb_t x, const arf_t y, slong prec) noexcept
    void arb_submul_si(arb_t z, const arb_t x, slong y, slong prec) noexcept
    void arb_submul_ui(arb_t z, const arb_t x, ulong y, slong prec) noexcept
    void arb_submul_fmpz(arb_t z, const arb_t x, const fmpz_t y, slong prec) noexcept
    void arb_fma(arb_t res, const arb_t x, const arb_t y, const arb_t z, slong prec) noexcept
    void arb_fma_arf(arb_t res, const arb_t x, const arf_t y, const arb_t z, slong prec) noexcept
    void arb_fma_si(arb_t res, const arb_t x, slong y, const arb_t z, slong prec) noexcept
    void arb_fma_ui(arb_t res, const arb_t x, ulong y, const arb_t z, slong prec) noexcept
    void arb_fma_fmpz(arb_t res, const arb_t x, const fmpz_t y, const arb_t z, slong prec) noexcept
    void arb_inv(arb_t z, const arb_t x, slong prec) noexcept
    void arb_div(arb_t z, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_div_arf(arb_t z, const arb_t x, const arf_t y, slong prec) noexcept
    void arb_div_si(arb_t z, const arb_t x, slong y, slong prec) noexcept
    void arb_div_ui(arb_t z, const arb_t x, ulong y, slong prec) noexcept
    void arb_div_fmpz(arb_t z, const arb_t x, const fmpz_t y, slong prec) noexcept
    void arb_fmpz_div_fmpz(arb_t z, const fmpz_t x, const fmpz_t y, slong prec) noexcept
    void arb_ui_div(arb_t z, ulong x, const arb_t y, slong prec) noexcept
    void arb_div_2expm1_ui(arb_t z, const arb_t x, ulong n, slong prec) noexcept
    void arb_dot_precise(arb_t res, const arb_t s, int subtract, arb_srcptr x, slong xstep, arb_srcptr y, slong ystep, slong len, slong prec) noexcept
    void arb_dot_simple(arb_t res, const arb_t s, int subtract, arb_srcptr x, slong xstep, arb_srcptr y, slong ystep, slong len, slong prec) noexcept
    void arb_dot(arb_t res, const arb_t s, int subtract, arb_srcptr x, slong xstep, arb_srcptr y, slong ystep, slong len, slong prec) noexcept
    void arb_approx_dot(arb_t res, const arb_t s, int subtract, arb_srcptr x, slong xstep, arb_srcptr y, slong ystep, slong len, slong prec) noexcept
    void arb_dot_ui(arb_t res, const arb_t initial, int subtract, arb_srcptr x, slong xstep, const ulong * y, slong ystep, slong len, slong prec) noexcept
    void arb_dot_si(arb_t res, const arb_t initial, int subtract, arb_srcptr x, slong xstep, const slong * y, slong ystep, slong len, slong prec) noexcept
    void arb_dot_uiui(arb_t res, const arb_t initial, int subtract, arb_srcptr x, slong xstep, const ulong * y, slong ystep, slong len, slong prec) noexcept
    void arb_dot_siui(arb_t res, const arb_t initial, int subtract, arb_srcptr x, slong xstep, const ulong * y, slong ystep, slong len, slong prec) noexcept
    void arb_dot_fmpz(arb_t res, const arb_t initial, int subtract, arb_srcptr x, slong xstep, const fmpz * y, slong ystep, slong len, slong prec) noexcept
    void arb_sqrt(arb_t z, const arb_t x, slong prec) noexcept
    void arb_sqrt_arf(arb_t z, const arf_t x, slong prec) noexcept
    void arb_sqrt_fmpz(arb_t z, const fmpz_t x, slong prec) noexcept
    void arb_sqrt_ui(arb_t z, ulong x, slong prec) noexcept
    void arb_sqrtpos(arb_t z, const arb_t x, slong prec) noexcept
    void arb_hypot(arb_t z, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_rsqrt(arb_t z, const arb_t x, slong prec) noexcept
    void arb_rsqrt_ui(arb_t z, ulong x, slong prec) noexcept
    void arb_sqrt1pm1(arb_t z, const arb_t x, slong prec) noexcept
    void arb_root_ui(arb_t z, const arb_t x, ulong k, slong prec) noexcept
    void arb_root(arb_t z, const arb_t x, ulong k, slong prec) noexcept
    void arb_sqr(arb_t y, const arb_t x, slong prec) noexcept
    void arb_pow_fmpz_binexp(arb_t y, const arb_t b, const fmpz_t e, slong prec) noexcept
    void arb_pow_fmpz(arb_t y, const arb_t b, const fmpz_t e, slong prec) noexcept
    void arb_pow_ui(arb_t y, const arb_t b, ulong e, slong prec) noexcept
    void arb_ui_pow_ui(arb_t y, ulong b, ulong e, slong prec) noexcept
    void arb_si_pow_ui(arb_t y, slong b, ulong e, slong prec) noexcept
    void arb_pow_fmpq(arb_t y, const arb_t x, const fmpq_t a, slong prec) noexcept
    void arb_pow(arb_t z, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_log_ui(arb_t z, ulong x, slong prec) noexcept
    void arb_log_fmpz(arb_t z, const fmpz_t x, slong prec) noexcept
    void arb_log_arf(arb_t z, const arf_t x, slong prec) noexcept
    void arb_log(arb_t z, const arb_t x, slong prec) noexcept
    void arb_log_ui_from_prev(arb_t log_k1, ulong k1, arb_t log_k0, ulong k0, slong prec) noexcept
    void arb_log1p(arb_t z, const arb_t x, slong prec) noexcept
    void arb_log_base_ui(arb_t res, const arb_t x, ulong b, slong prec) noexcept
    void arb_log_hypot(arb_t res, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_exp(arb_t z, const arb_t x, slong prec) noexcept
    void arb_expm1(arb_t z, const arb_t x, slong prec) noexcept
    void arb_exp_invexp(arb_t z, arb_t w, const arb_t x, slong prec) noexcept
    void arb_sin(arb_t s, const arb_t x, slong prec) noexcept
    void arb_cos(arb_t c, const arb_t x, slong prec) noexcept
    void arb_sin_cos(arb_t s, arb_t c, const arb_t x, slong prec) noexcept
    void arb_sin_pi(arb_t s, const arb_t x, slong prec) noexcept
    void arb_cos_pi(arb_t c, const arb_t x, slong prec) noexcept
    void arb_sin_cos_pi(arb_t s, arb_t c, const arb_t x, slong prec) noexcept
    void arb_tan(arb_t y, const arb_t x, slong prec) noexcept
    void arb_cot(arb_t y, const arb_t x, slong prec) noexcept
    void arb_sin_cos_pi_fmpq(arb_t s, arb_t c, const fmpq_t x, slong prec) noexcept
    void arb_sin_pi_fmpq(arb_t s, const fmpq_t x, slong prec) noexcept
    void arb_cos_pi_fmpq(arb_t c, const fmpq_t x, slong prec) noexcept
    void arb_tan_pi(arb_t y, const arb_t x, slong prec) noexcept
    void arb_cot_pi(arb_t y, const arb_t x, slong prec) noexcept
    void arb_sec(arb_t res, const arb_t x, slong prec) noexcept
    void arb_csc(arb_t res, const arb_t x, slong prec) noexcept
    void arb_csc_pi(arb_t res, const arb_t x, slong prec) noexcept
    void arb_sinc(arb_t z, const arb_t x, slong prec) noexcept
    void arb_sinc_pi(arb_t z, const arb_t x, slong prec) noexcept
    void arb_atan_arf(arb_t z, const arf_t x, slong prec) noexcept
    void arb_atan(arb_t z, const arb_t x, slong prec) noexcept
    void arb_atan2(arb_t z, const arb_t b, const arb_t a, slong prec) noexcept
    void arb_asin(arb_t z, const arb_t x, slong prec) noexcept
    void arb_acos(arb_t z, const arb_t x, slong prec) noexcept
    void arb_sinh(arb_t s, const arb_t x, slong prec) noexcept
    void arb_cosh(arb_t c, const arb_t x, slong prec) noexcept
    void arb_sinh_cosh(arb_t s, arb_t c, const arb_t x, slong prec) noexcept
    void arb_tanh(arb_t y, const arb_t x, slong prec) noexcept
    void arb_coth(arb_t y, const arb_t x, slong prec) noexcept
    void arb_sech(arb_t res, const arb_t x, slong prec) noexcept
    void arb_csch(arb_t res, const arb_t x, slong prec) noexcept
    void arb_atanh(arb_t z, const arb_t x, slong prec) noexcept
    void arb_asinh(arb_t z, const arb_t x, slong prec) noexcept
    void arb_acosh(arb_t z, const arb_t x, slong prec) noexcept
    void arb_const_pi(arb_t z, slong prec) noexcept
    void arb_const_sqrt_pi(arb_t z, slong prec) noexcept
    void arb_const_log_sqrt2pi(arb_t z, slong prec) noexcept
    void arb_const_log2(arb_t z, slong prec) noexcept
    void arb_const_log10(arb_t z, slong prec) noexcept
    void arb_const_euler(arb_t z, slong prec) noexcept
    void arb_const_catalan(arb_t z, slong prec) noexcept
    void arb_const_e(arb_t z, slong prec) noexcept
    void arb_const_khinchin(arb_t z, slong prec) noexcept
    void arb_const_glaisher(arb_t z, slong prec) noexcept
    void arb_const_apery(arb_t z, slong prec) noexcept
    void arb_const_reciprocal_fibonacci(arb_t z, slong prec) noexcept
    void arb_lambertw(arb_t res, const arb_t x, int flags, slong prec) noexcept
    void arb_rising_ui(arb_t z, const arb_t x, ulong n, slong prec) noexcept
    void arb_rising(arb_t z, const arb_t x, const arb_t n, slong prec) noexcept
    void arb_rising_fmpq_ui(arb_t z, const fmpq_t x, ulong n, slong prec) noexcept
    void arb_fac_ui(arb_t z, ulong n, slong prec) noexcept
    void arb_doublefac_ui(arb_t z, ulong n, slong prec) noexcept
    void arb_bin_ui(arb_t z, const arb_t n, ulong k, slong prec) noexcept
    void arb_bin_uiui(arb_t z, ulong n, ulong k, slong prec) noexcept
    void arb_gamma(arb_t z, const arb_t x, slong prec) noexcept
    void arb_gamma_fmpq(arb_t z, const fmpq_t x, slong prec) noexcept
    void arb_gamma_fmpz(arb_t z, const fmpz_t x, slong prec) noexcept
    void arb_lgamma(arb_t z, const arb_t x, slong prec) noexcept
    void arb_rgamma(arb_t z, const arb_t x, slong prec) noexcept
    void arb_digamma(arb_t y, const arb_t x, slong prec) noexcept
    void arb_zeta_ui_vec_borwein(arb_ptr z, ulong start, slong num, ulong step, slong prec) noexcept
    void arb_zeta_ui_asymp(arb_t x, ulong s, slong prec) noexcept
    void arb_zeta_ui_euler_product(arb_t z, ulong s, slong prec) noexcept
    void arb_zeta_ui_bernoulli(arb_t x, ulong s, slong prec) noexcept
    void arb_zeta_ui_borwein_bsplit(arb_t x, ulong s, slong prec) noexcept
    void arb_zeta_ui_vec(arb_ptr x, ulong start, slong num, slong prec) noexcept
    void arb_zeta_ui_vec_even(arb_ptr x, ulong start, slong num, slong prec) noexcept
    void arb_zeta_ui_vec_odd(arb_ptr x, ulong start, slong num, slong prec) noexcept
    void arb_zeta_ui(arb_t x, ulong s, slong prec) noexcept
    void arb_zeta(arb_t z, const arb_t s, slong prec) noexcept
    void arb_hurwitz_zeta(arb_t z, const arb_t s, const arb_t a, slong prec) noexcept
    void arb_bernoulli_ui(arb_t b, ulong n, slong prec) noexcept
    void arb_bernoulli_fmpz(arb_t b, const fmpz_t n, slong prec) noexcept
    void arb_bernoulli_ui_zeta(arb_t b, ulong n, slong prec) noexcept
    void arb_bernoulli_poly_ui(arb_t res, ulong n, const arb_t x, slong prec) noexcept
    void arb_power_sum_vec(arb_ptr res, const arb_t a, const arb_t b, slong len, slong prec) noexcept
    void arb_polylog(arb_t w, const arb_t s, const arb_t z, slong prec) noexcept
    void arb_polylog_si(arb_t w, slong s, const arb_t z, slong prec) noexcept
    void arb_fib_fmpz(arb_t z, const fmpz_t n, slong prec) noexcept
    void arb_fib_ui(arb_t z, ulong n, slong prec) noexcept
    void arb_agm(arb_t z, const arb_t x, const arb_t y, slong prec) noexcept
    void arb_chebyshev_t_ui(arb_t a, ulong n, const arb_t x, slong prec) noexcept
    void arb_chebyshev_u_ui(arb_t a, ulong n, const arb_t x, slong prec) noexcept
    void arb_chebyshev_t2_ui(arb_t a, arb_t b, ulong n, const arb_t x, slong prec) noexcept
    void arb_chebyshev_u2_ui(arb_t a, arb_t b, ulong n, const arb_t x, slong prec) noexcept
    void arb_bell_sum_bsplit(arb_t res, const fmpz_t n, const fmpz_t a, const fmpz_t b, const fmpz_t mmag, slong prec) noexcept
    void arb_bell_sum_taylor(arb_t res, const fmpz_t n, const fmpz_t a, const fmpz_t b, const fmpz_t mmag, slong prec) noexcept
    void arb_bell_fmpz(arb_t res, const fmpz_t n, slong prec) noexcept
    void arb_bell_ui(arb_t res, ulong n, slong prec) noexcept
    void arb_euler_number_fmpz(arb_t res, const fmpz_t n, slong prec) noexcept
    void arb_euler_number_ui(arb_t res, ulong n, slong prec) noexcept
    void arb_fmpz_euler_number_ui_multi_mod(fmpz_t res, ulong n, double alpha) noexcept
    void arb_fmpz_euler_number_ui(fmpz_t res, ulong n) noexcept
    void arb_partitions_fmpz(arb_t res, const fmpz_t n, slong prec) noexcept
    void arb_partitions_ui(arb_t res, ulong n, slong prec) noexcept
    void arb_primorial_nth_ui(arb_t res, ulong n, slong prec) noexcept
    void arb_primorial_ui(arb_t res, ulong n, slong prec) noexcept
    void _arb_atan_taylor_naive(mp_ptr y, mp_limb_t * error, mp_srcptr x, mp_size_t xn, ulong N, int alternating) noexcept
    void _arb_atan_taylor_rs(mp_ptr y, mp_limb_t * error, mp_srcptr x, mp_size_t xn, ulong N, int alternating) noexcept
    void _arb_exp_taylor_naive(mp_ptr y, mp_limb_t * error, mp_srcptr x, mp_size_t xn, ulong N) noexcept
    void _arb_exp_taylor_rs(mp_ptr y, mp_limb_t * error, mp_srcptr x, mp_size_t xn, ulong N) noexcept
    void _arb_sin_cos_taylor_naive(mp_ptr ysin, mp_ptr ycos, mp_limb_t * error, mp_srcptr x, mp_size_t xn, ulong N) noexcept
    void _arb_sin_cos_taylor_rs(mp_ptr ysin, mp_ptr ycos, mp_limb_t * error, mp_srcptr x, mp_size_t xn, ulong N, int sinonly, int alternating) noexcept
    int _arb_get_mpn_fixed_mod_log2(mp_ptr w, fmpz_t q, mp_limb_t * error, const arf_t x, mp_size_t wn) noexcept
    int _arb_get_mpn_fixed_mod_pi4(mp_ptr w, fmpz_t q, int * octant, mp_limb_t * error, const arf_t x, mp_size_t wn) noexcept
    slong _arb_exp_taylor_bound(slong mag, slong prec) noexcept
    void arb_exp_arf_bb(arb_t z, const arf_t x, slong prec, int m1) noexcept
    void _arb_exp_sum_bs_simple(fmpz_t T, fmpz_t Q, flint_bitcnt_t * Qexp, const fmpz_t x, flint_bitcnt_t r, slong N) noexcept
    void _arb_exp_sum_bs_powtab(fmpz_t T, fmpz_t Q, flint_bitcnt_t * Qexp, const fmpz_t x, flint_bitcnt_t r, slong N) noexcept
    void arb_exp_arf_rs_generic(arb_t res, const arf_t x, slong prec, int minus_one) noexcept
    void _arb_atan_sum_bs_simple(fmpz_t T, fmpz_t Q, flint_bitcnt_t * Qexp, const fmpz_t x, flint_bitcnt_t r, slong N) noexcept
    void _arb_atan_sum_bs_powtab(fmpz_t T, fmpz_t Q, flint_bitcnt_t * Qexp, const fmpz_t x, flint_bitcnt_t r, slong N) noexcept
    void arb_atan_arf_bb(arb_t z, const arf_t x, slong prec) noexcept
    void arb_atan_frac_bsplit(arb_t s, const fmpz_t p, const fmpz_t q, int hyperbolic, slong prec) noexcept
    void arb_sin_cos_arf_generic(arb_t s, arb_t c, const arf_t x, slong prec) noexcept
    void arb_sin_cos_arf_bb(arb_t s, arb_t c, const arf_t x, slong prec) noexcept
    void arb_sin_cos_wide(arb_t s, arb_t c, const arb_t x, slong prec) noexcept
    void arb_sin_cos_generic(arb_t s, arb_t c, const arb_t x, slong prec) noexcept
    void arb_log_primes_vec_bsplit(arb_ptr res, slong n, slong prec) noexcept
    void _arb_log_p_ensure_cached(slong prec) noexcept
    void arb_exp_arf_log_reduction(arb_t res, const arf_t x, slong prec, int minus_one) noexcept
    void arb_exp_arf_generic(arb_t z, const arf_t x, slong prec, int minus_one) noexcept
    void arb_exp_arf(arb_t z, const arf_t x, slong prec, int minus_one, slong maglim) noexcept
    void arb_log_newton(arb_t res, const arb_t x, slong prec) noexcept
    void arb_log_arf_newton(arb_t res, const arf_t x, slong prec) noexcept
    void arb_atan_gauss_primes_vec_bsplit(arb_ptr res, slong n, slong prec) noexcept
    void _arb_atan_gauss_p_ensure_cached(slong prec) noexcept
    void arb_sin_cos_arf_atan_reduction(arb_t res1, arb_t res2, const arf_t x, slong prec) noexcept
    void arb_atan_newton(arb_t res, const arb_t x, slong prec) noexcept
    void arb_atan_arf_newton(arb_t res, const arf_t x, slong prec) noexcept
    void _arb_vec_zero(arb_ptr vec, slong n) noexcept
    bint _arb_vec_is_zero(arb_srcptr vec, slong len) noexcept
    bint _arb_vec_is_finite(arb_srcptr x, slong len) noexcept
    bint _arb_vec_equal(arb_srcptr vec1, arb_srcptr vec2, slong len) noexcept
    bint _arb_vec_overlaps(arb_srcptr vec1, arb_srcptr vec2, slong len) noexcept
    bint _arb_vec_contains(arb_srcptr vec1, arb_srcptr vec2, slong len) noexcept
    void _arb_vec_set(arb_ptr res, arb_srcptr vec, slong len) noexcept
    void _arb_vec_set_round(arb_ptr res, arb_srcptr vec, slong len, slong prec) noexcept
    void _arb_vec_swap(arb_ptr vec1, arb_ptr vec2, slong len) noexcept
    void _arb_vec_neg(arb_ptr B, arb_srcptr A, slong n) noexcept
    void _arb_vec_sub(arb_ptr C, arb_srcptr A, arb_srcptr B, slong n, slong prec) noexcept
    void _arb_vec_add(arb_ptr C, arb_srcptr A, arb_srcptr B, slong n, slong prec) noexcept
    void _arb_vec_scalar_mul(arb_ptr res, arb_srcptr vec, slong len, const arb_t c, slong prec) noexcept
    void _arb_vec_scalar_div(arb_ptr res, arb_srcptr vec, slong len, const arb_t c, slong prec) noexcept
    void _arb_vec_scalar_mul_fmpz(arb_ptr res, arb_srcptr vec, slong len, const fmpz_t c, slong prec) noexcept
    void _arb_vec_scalar_mul_2exp_si(arb_ptr res, arb_srcptr src, slong len, slong c) noexcept
    void _arb_vec_scalar_addmul(arb_ptr res, arb_srcptr vec, slong len, const arb_t c, slong prec) noexcept
    void _arb_vec_get_mag(mag_t bound, arb_srcptr vec, slong len) noexcept
    slong _arb_vec_bits(arb_srcptr x, slong len) noexcept
    void _arb_vec_set_powers(arb_ptr xs, const arb_t x, slong len, slong prec) noexcept
    void _arb_vec_add_error_arf_vec(arb_ptr res, arf_srcptr err, slong len) noexcept
    void _arb_vec_add_error_mag_vec(arb_ptr res, mag_srcptr err, slong len) noexcept
    void _arb_vec_indeterminate(arb_ptr vec, slong len) noexcept
    void _arb_vec_trim(arb_ptr res, arb_srcptr vec, slong len) noexcept
    int _arb_vec_get_unique_fmpz_vec(fmpz * res,  arb_srcptr vec, slong len) noexcept
    void _arb_vec_printn(arb_srcptr vec, slong len, slong digits, ulong flags) noexcept
    void _arb_vec_printd(arb_srcptr vec, slong len, slong ndigits) noexcept

from .arb_macros cimport *



r"""
Logic Tables

A logic table is essentially a 2-D array that is created by the statement class
and stored in the private global variable table, along with a list containing
the variable names to be used, in order.

The order in which the table is listed essentially amounts to counting in
binary. For instance, with the variables `A`, `B`, and `C` the truth table
looks like::

    A      B      C      value
    False  False  False    ?
    False  False  True     ?
    False  True   False    ?
    False  True   True     ?
    True   False  False    ?
    True   False  True     ?
    True   True   False    ?
    True   True   True     ?

This is equivalent to counting in binary, where a table would appear thus;

::

    2^2 2^1 2^0 value
    0   0   0     0
    0   0   1     1
    0   1   0     2
    0   1   1     3
    1   0   0     4
    1   0   1     5
    1   1   0     6
    1   1   1     7

Given that a table can be created corresponding to any range of acceptable
values for a given statement, it is easy to find the value of a statement
for arbitrary values of its variables.

AUTHORS:

- William Stein (2006): initial version

- Chris Gorecki (2006): initial version

- Paul Scurek (2013-08-03): updated docstring formatting

EXAMPLES:

Create a truth table of a boolean formula::

    sage: import sage.logic.propcalc as propcalc
    sage: s = propcalc.formula("a&b|~(c|a)")
    sage: s.truthtable()
    a      b      c      value
    False  False  False  True
    False  False  True   False
    False  True   False  True
    False  True   True   False
    True   False  False  False
    True   False  True   False
    True   True   False  True
    True   True   True   True

Get the LaTeX code for a truth table::

    sage: latex(s.truthtable(5,11))
    \\\begin{tabular}{llll}c & b & a & value \\\hline True & False & True & False \\True & True & False & True \\True & True & True & True\end{tabular}

It is not an error to use nonsensical numeric inputs::

    sage: s = propcalc.formula("a&b|~(c|a)")
    sage: s.truthtable(5, 9)
    a      b      c      value
    True   False  True   False
    True   True   False  True
    True   True   True   True

    sage: s.truthtable(9, 5)
    a      b      c      value

If one argument is provided, truthtable defaults to the end::

    sage: s.truthtable(-1)
    a      b      c      value
    False  False  False  True
    False  False  True   False
    False  True   False  True
    False  True   True   False
    True   False  False  False
    True   False  True   False
    True   True   False  True
    True   True   True   True

If the second argument is negative, truthtable defaults to the end::

    sage: s.truthtable(4, -2)
    a      b      c      value
    True   False  False  False
    True   False  True   False
    True   True   False  True
    True   True   True   True

.. NOTE::

    For statements that contain a variable list that when printed is longer
    than the latex page, the columns of the table will run off the screen.
"""
# ****************************************************************************
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#       Copyright (C) 2006 Chris Gorecki <chris.k.gorecki@gmail.com>
#       Copyright (C) 2013 Paul Scurek <scurek86@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

# Global variables
__table = []
__vars_order = []


class Truthtable:
    """
    A truth table.

    INPUT:

    - ``t`` -- a 2-D array containing the table values

    - ``vo`` -- list of the variables in the expression in order,
      with each variable occurring only once
    """
    def __init__(self, t, vo):
        r"""
        Initialize the data fields.

        EXAMPLES:

        This example illustrates the creation of a table

        ::

            sage: import sage.logic.propcalc as propcalc
            sage: s = propcalc.formula("a&b|~(c|a)")
            sage: s.truthtable()
            a      b      c      value
            False  False  False  True
            False  False  True   False
            False  True   False  True
            False  True   True   False
            True   False  False  False
            True   False  True   False
            True   True   False  True
            True   True   True   True
        """
        self.__table = t
        self.__vars_order = vo

    def _latex_(self):
        r"""
        Return a latex representation of the calling table object.

        OUTPUT: the latex representation of the table

        EXAMPLES:

        This example illustrates how to get the latex representation of
        the table::

            sage: import sage.logic.propcalc as propcalc
            sage: s = propcalc.formula("man->monkey&human")
            sage: latex(s.truthtable())
            \\\begin{tabular}{llll}human & monkey & man & value \\\hline False & False & False & True \\False & False & True & True \\False & True & False & True \\False & True & True & True \\True & False & False & False \\True & False & True & False \\True & True & False & False \\True & True & True & True\end{tabular}

        Now, we show that strange parameters can lead to a table header
        with no body::

            sage: latex(s.truthtable(2, 1))
            \\\begin{tabular}{llll}human & monkey & man & value \\\hli\end{tabular}
        """
        rt = s = ""
        self.__vars_order.reverse()
        s = r'\\\begin{tabular}{'
        s += 'l' * (len(self.__vars_order) + 1) + '}'
        for var in self.__vars_order:
            s += var + ' & '
        rt += s + r'value \\' + r'\hline '
        for row in self.__table:
            s = ""
            for i in row:
                s += str(i) + ' & '
            rt += s[:-3] + r' \\'
        rt = rt[:-3] + r'\end{tabular}'
        self.__vars_order.reverse()
        return rt

    def __repr__(self):
        r"""
        Return a string representation of the calling table object.

        OUTPUT: the table as a 2-D string array

        EXAMPLES:

        This example illustrates how to display the truth table of a
        boolean formula::

            sage: import sage.logic.propcalc as propcalc
            sage: s = propcalc.formula("man->monkey&human")
            sage: s.truthtable()
            man    monkey  human  value
            False  False   False  True
            False  False   True   True
            False  True    False  True
            False  True    True   True
            True   False   False  False
            True   False   True   False
            True   True    False  False
            True   True    True   True

        We now show that strange parameters can lead to the table
        header with no body::

            sage: s.truthtable(2, 1)
            man    monkey  human  value
        """
        vars_len = []
        line = rt = s = ""
        for var in self.__vars_order:
            vars_len.append(len(var))
            s = var + ' '
            while len(s) < len('False '):
                s += ' '
            s += ' '
            line += s
        rt += line + 'value\n'
        for row in self.__table:
            line = s = ""
            i = 0
            for e in row:
                j = 2 if e else 1
                s = str(e) + ' ' * j
                if i < len(vars_len):
                    while len(s) <= vars_len[i]:
                        s += ' '
                s += ' '
                line += s
                i += 1
            rt += line + '\n'
        return rt

    def get_table_list(self):
        r"""
        Return a list representation of the calling table object.

        OUTPUT: list representation of the table

        EXAMPLES:

        This example illustrates how to show the table as a list::

            sage: import sage.logic.propcalc as propcalc
            sage: s = propcalc.formula("man->monkey&human")
            sage: s.truthtable().get_table_list()
             [['man', 'monkey', 'human'], [False, False, False, True], [False, False, True, True], [False, True, False, True], [False, True, True, True], [True, False, False, False], [True, False, True, False], [True, True, False, False], [True, True, True, True]]
        """
        t = self.__table[:]
        t.insert(0, self.__vars_order)
        return t



# sage_setup: distribution = sagemath-modules
# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
from sage.libs.gsl.types cimport *

cdef extern from "gsl/gsl_sf_gamma.h":

  double  gsl_sf_gamma(double x)

  int  gsl_sf_gamma_e(double x, gsl_sf_result * result)

  double  gsl_sf_lngamma(double x)

  int  gsl_sf_lngamma_e(double x, gsl_sf_result * result)

  int  gsl_sf_lngamma_sgn_e(double x, gsl_sf_result * result_lg, double * sgn)

  double  gsl_sf_gammastar(double x)

  int  gsl_sf_gammastar_e(double x, gsl_sf_result * result)

  double  gsl_sf_gammainv(double x)

  int  gsl_sf_gammainv_e(double x, gsl_sf_result * result)

  int  gsl_sf_lngamma_complex_e(double zr, double zi, gsl_sf_result * lnr, gsl_sf_result * arg)

  double  gsl_sf_taylorcoeff(int n, double x)

  int  gsl_sf_taylorcoeff_e(int n, double x, gsl_sf_result * result)

  double  gsl_sf_fact(unsigned int n)

  int  gsl_sf_fact_e(unsigned int n, gsl_sf_result * result)

  double  gsl_sf_doublefact(unsigned int n)

  int  gsl_sf_doublefact_e(unsigned int n, gsl_sf_result * result)

  double  gsl_sf_lnfact(unsigned int n)

  int  gsl_sf_lnfact_e(unsigned int n, gsl_sf_result * result)

  double  gsl_sf_lndoublefact(unsigned int n)

  int  gsl_sf_lndoublefact_e(unsigned int n, gsl_sf_result * result)

  double  gsl_sf_choose(unsigned int n, unsigned int m)

  int  gsl_sf_choose_e(unsigned int n, unsigned int m, gsl_sf_result * result)

  double  gsl_sf_lnchoose(unsigned int n, unsigned int m)

  int  gsl_sf_lnchoose_e(unsigned int n, unsigned int m, gsl_sf_result * result)

  double  gsl_sf_poch(double a, double x)

  int  gsl_sf_poch_e(double a, double x, gsl_sf_result * result)

  double  gsl_sf_lnpoch(double a, double x)

  int  gsl_sf_lnpoch_e(double a, double x, gsl_sf_result * result)

  int  gsl_sf_lnpoch_sgn_e(double a, double x, gsl_sf_result * result, double * sgn)

  double  gsl_sf_pochrel(double a, double x)

  int  gsl_sf_pochrel_e(double a, double x, gsl_sf_result * result)

  double  gsl_sf_gamma_inc_Q(double a, double x)

  int  gsl_sf_gamma_inc_Q_e(double a, double x, gsl_sf_result * result)

  double  gsl_sf_gamma_inc_P(double a, double x)

  int  gsl_sf_gamma_inc_P_e(double a, double x, gsl_sf_result * result)

  double  gsl_sf_gamma_inc(double a, double x)

  int  gsl_sf_gamma_inc_e(double a, double x, gsl_sf_result * result)

  double  gsl_sf_beta(double a, double b)

  int  gsl_sf_beta_e(double a, double b, gsl_sf_result * result)

  double  gsl_sf_lnbeta(double a, double b)

  int  gsl_sf_lnbeta_e(double a, double b, gsl_sf_result * result)

  double  gsl_sf_beta_inc(double a, double b, double x)

  int  gsl_sf_beta_inc_e(double a, double b, double x, gsl_sf_result * result)




# sage_setup: distribution = sagemath-flint
# distutils: extra_compile_args = -D_XPG6

# WARNING: src/sage/libs/flint/flint_sage.pyx is generated from
# src/sage_setup/autogen/flint/templates/flint_sage.pyx.template;
# please make sure that you are modifying the correct file!
"""
Flint imports

TESTS:

Import this module::

    sage: import sage.libs.flint.flint_sage

We verify that :issue:`6919` is correctly fixed::

    sage: R.<x> = PolynomialRing(ZZ)
    sage: A = 2^(2^17+2^15)
    sage: a = A * x^31
    sage: b = (A * x) * x^30
    sage: a == b
    True
"""

# cimport all .pxd files to make sure they compile
from .acb cimport *
from .acb_calc cimport *
from .acb_dft cimport *
from .acb_dirichlet cimport *
from .acb_elliptic cimport *
from .acb_hypgeom cimport *
from .acb_mat cimport *
from .acb_modular cimport *
from .acb_poly cimport *
from .acf cimport *
from .aprcl cimport *
from .arb cimport *
from .arb_calc cimport *
from .arb_fmpz_poly cimport *
from .arb_fpwrap cimport *
from .arb_hypgeom cimport *
from .arb_mat cimport *
from .arb_poly cimport *
from .arf cimport *
from .arith cimport *
from .bernoulli cimport *
from .bool_mat cimport *
from .ca cimport *
from .ca_ext cimport *
from .ca_field cimport *
from .ca_mat cimport *
from .ca_poly cimport *
from .ca_vec cimport *
from .calcium cimport *
from .d_mat cimport *
from .d_vec cimport *
from .dirichlet cimport *
from .dlog cimport *
from .double_extras cimport *
from .double_interval cimport *
from .fexpr cimport *
from .fexpr_builtin cimport *
from .fft cimport *
from .flint cimport *
from .fmpq cimport *
from .fmpq_mat cimport *
from .fmpq_mpoly cimport *
from .fmpq_mpoly_factor cimport *
from .fmpq_poly cimport *
from .fmpq_vec cimport *
from .fmpz cimport *
from .fmpz_extras cimport *
from .fmpz_factor cimport *
from .fmpz_lll cimport *
from .fmpz_mat cimport *
from .fmpz_mod cimport *
from .fmpz_mod_mat cimport *
from .fmpz_mod_mpoly cimport *
from .fmpz_mod_mpoly_factor cimport *
from .fmpz_mod_poly cimport *
from .fmpz_mod_poly_factor cimport *
from .fmpz_mod_vec cimport *
from .fmpz_mpoly cimport *
from .fmpz_mpoly_factor cimport *
from .fmpz_mpoly_q cimport *
from .fmpz_poly cimport *
from .fmpz_poly_factor cimport *
from .fmpz_poly_mat cimport *
from .fmpz_poly_q cimport *
from .fmpz_vec cimport *
from .fmpzi cimport *
from .fq cimport *
from .fq_default cimport *
from .fq_default_mat cimport *
from .fq_default_poly cimport *
from .fq_default_poly_factor cimport *
from .fq_embed cimport *
from .fq_mat cimport *
from .fq_nmod cimport *
from .fq_nmod_embed cimport *
from .fq_nmod_mat cimport *
from .fq_nmod_mpoly cimport *
from .fq_nmod_mpoly_factor cimport *
from .fq_nmod_poly cimport *
from .fq_nmod_poly_factor cimport *
from .fq_nmod_vec cimport *
from .fq_poly cimport *
from .fq_poly_factor cimport *
from .fq_vec cimport *
from .fq_zech cimport *
from .fq_zech_embed cimport *
from .fq_zech_mat cimport *
from .fq_zech_poly cimport *
from .fq_zech_poly_factor cimport *
from .fq_zech_vec cimport *
from .gr cimport *
from .gr_generic cimport *
from .gr_mat cimport *
from .gr_mpoly cimport *
from .gr_poly cimport *
from .gr_special cimport *
from .gr_vec cimport *
from .hypgeom cimport *
from .long_extras cimport *
from .mag cimport *
from .mpf_mat cimport *
from .mpf_vec cimport *
from .mpfr_mat cimport *
from .mpfr_vec cimport *
from .mpn_extras cimport *
from .mpoly cimport *
from .nf cimport *
from .nf_elem cimport *
from .nmod cimport *
from .nmod_mat cimport *
from .nmod_mpoly cimport *
from .nmod_mpoly_factor cimport *
from .nmod_poly cimport *
from .nmod_poly_factor cimport *
from .nmod_poly_mat cimport *
from .nmod_vec cimport *
from .padic cimport *
from .padic_mat cimport *
from .padic_poly cimport *
from .partitions cimport *
from .perm cimport *
from .profiler cimport *
from .qadic cimport *
from .qfb cimport *
from .qqbar cimport *
from .qsieve cimport *
from .thread_pool cimport *
from .ulong_extras cimport *

# Try to clean up after ourselves before sage terminates. This
# probably doesn't do anything if your copy of flint is re-entrant
# (and most are). Moreover it isn't strictly necessary, because the OS
# will reclaim these resources anyway after sage terminates. However
# this might reveal other bugs, and can help tools like valgrind do
# their jobs.
import atexit
atexit.register(_fmpz_cleanup_mpz_content)



# sage_setup: distribution = sagemath-gap
# distutils: libraries = gap gmp M_LIBRARIES
###############################################################################
#       Copyright (C) 2009, William Stein <wstein@gmail.com>
#       Copyright (C) 2012, Volker Braun <vbraun.name@gmail.com>
#
#   Distributed under the terms of the GNU General Public License (GPL)
#   as published by the Free Software Foundation; either version 2 of
#   the License, or (at your option) any later version.
#                   http://www.gnu.org/licenses/
###############################################################################

from libc.stdint cimport intptr_t, uintptr_t, uint8_t, uint16_t, uint32_t, uint64_t

cdef extern from "gap/system.h" nogil:
    ctypedef char Char
    ctypedef intptr_t Int
    ctypedef uintptr_t UInt
    ctypedef uint8_t  UInt1
    ctypedef uint16_t UInt2
    ctypedef uint32_t UInt4
    ctypedef uint64_t UInt8
    ctypedef void* Obj


cdef extern from "gap/calls.h" nogil:
    bint IS_FUNC(Obj)


cdef extern from "gap/libgap-api.h" nogil:
    ctypedef void (*GAP_CallbackFunc)()
    void GAP_Initialize(int argc, char ** argv,
            GAP_CallbackFunc markBagsCallback, GAP_CallbackFunc errorCallback,
            int handleSignals)
    Obj GAP_EvalString(const char *) except *
    Obj GAP_EvalStringNoExcept "GAP_EvalString"(const char *)
    Obj GAP_ValueGlobalVariable(const char *)
    cdef void GAP_EnterStack()
    cdef void GAP_LeaveStack()
    cdef int GAP_Enter() except 0
    cdef void GAP_Leave()
    cdef int GAP_Error_Setjmp() except 0

    void GAP_MarkBag(Obj bag)
    void GAP_CollectBags(UInt full)

    Obj GAP_SUM(Obj, Obj)
    Obj GAP_DIFF(Obj, Obj)
    Obj GAP_PROD(Obj, Obj)
    Obj GAP_QUO(Obj, Obj)
    Obj GAP_POW(Obj, Obj)
    Obj GAP_MOD(Obj, Obj)
    bint GAP_EQ(Obj opL, Obj opR)
    bint GAP_LT(Obj opL, Obj opR)
    bint GAP_IN(Obj opL, Obj opR)

    cdef Obj GAP_True
    cdef Obj GAP_False
    cdef Obj GAP_Fail

    Obj GAP_CallFuncList(Obj func, Obj args);
    Obj GAP_CallFuncArray(Obj func, UInt narg, Obj * args);
    Obj GAP_CallFunc0Args(Obj func);
    Obj GAP_CallFunc1Args(Obj func, Obj a1);
    Obj GAP_CallFunc2Args(Obj func, Obj a1, Obj a2);
    Obj GAP_CallFunc3Args(Obj func, Obj a1, Obj a2, Obj a3);

    bint GAP_IsMacFloat(Obj obj)
    double GAP_ValueMacFloat(Obj obj)

    bint GAP_IsInt(Obj)
    bint GAP_IsSmallInt(Obj)
    Obj GAP_NewObjIntFromInt(Int val)
    Obj GAP_MakeObjInt(const UInt* limbs, Int size)
    Int GAP_ValueInt(Obj)
    Int GAP_SizeInt(Obj)
    const UInt* GAP_AddrInt(Obj)

    bint GAP_IsList(Obj lst)
    UInt GAP_LenList(Obj lst)
    void GAP_AssList(Obj lst, UInt pos, Obj val)
    Obj GAP_ElmList(Obj lst, UInt pos)
    Obj GAP_NewPlist(Int capacity)

    bint GAP_IsRecord(Obj obj)
    Obj GAP_NewPrecord(Int capacity)

    bint GAP_IsString(Obj obj)
    UInt GAP_LenString(Obj string)
    char* GAP_CSTR_STRING(Obj list)
    Obj GAP_MakeStringWithLen(const char* buf, UInt len)

    bint GAP_IsChar(Obj obj)


cdef extern from "gap/lists.h" nogil:
    Obj ELM_LIST(Obj lst, int pos)


cdef extern from "gap/objects.h" nogil:
    bint IS_MUTABLE_OBJ(Obj obj)
    Obj SHALLOW_COPY_OBJ(Obj obj)
    Obj CopyObj(Obj obj, int mut)

    UInt TNUM_OBJ(Obj obj)
    const char* TNAM_OBJ(Obj obj)

    cdef enum TNUM:
        T_RAT
        T_CYC
        T_FFE
        T_PERM2
        T_PERM4
        T_BOOL
        T_FUNCTION
        T_COMOBJ
        T_POSOBJ


cdef extern from "gap/permutat.h" nogil:
    UInt DEG_PERM2(Obj)
    UInt DEG_PERM4(Obj)
    Obj NEW_PERM2(UInt)
    Obj NEW_PERM4(UInt)
    UInt2* ADDR_PERM2(Obj)
    UInt4* ADDR_PERM4(Obj)
    const UInt2* CONST_ADDR_PERM2(Obj)
    const UInt4* CONST_ADDR_PERM4(Obj)


cdef extern from "gap/precord.h" nogil:
    int LEN_PREC(Obj rec)
    int GET_RNAM_PREC(Obj rec, int i)
    Obj GET_ELM_PREC(Obj rec, int i)
    void AssPRec(Obj rec, UInt rnam, Obj val)


cdef extern from "gap/records.h" nogil:
    char* NAME_RNAM(UInt rnam)
    Obj ELM_REC(Obj rec, UInt rnam)
    UInt RNamName(Char* name)


cdef extern from "gap/stringobj.h" nogil:
    Obj NEW_STRING(Int)


cdef extern from "gap/stats.h" nogil:
    void InterruptExecStat()


cdef extern from "<structmember.h>" nogil:
    """
    /* Hack: Cython 3.0 automatically includes <structmember.h>, which
     * defines several macros that collides with enum definitions in
     * gap/objects.h. We need to include the header explicitly and
     * undefine these macros.
     */
    #undef T_INT
    #undef T_STRING
    #undef T_CHAR
    #undef T_BOOL
    """
    pass



# sage_setup: distribution = sagemath-modules
from sage.rings.polynomial.ore_polynomial_element cimport OrePolynomial_generic_dense

cdef class SkewPolynomial_generic_dense(OrePolynomial_generic_dense):
    cdef void _inplace_rmul(self, SkewPolynomial_generic_dense right) noexcept
    cdef void _inplace_pow(self, Py_ssize_t n) noexcept
    cpdef right_power_mod(self, exp, modulus)
    cpdef left_power_mod(self, exp, modulus)
    cpdef operator_eval(self, eval_pt)



# sage_setup: distribution = sagemath-flint
r"""
Deprecated module

TESTS::

    sage: from sage.libs.flint.fmpz_poly import Fmpz_poly
    sage: Fmpz_poly([1, 1])
    doctest:warning
    ...
    DeprecationWarning:
    Importing Fmpz_poly from here is deprecated; please use "from sage.libs.flint.fmpz_poly_sage import Fmpz_poly" instead.
    See https://github.com/sagemath/sage/issues/36449 for details.
    2  1 1
"""

from sage.misc.lazy_import import LazyImport

Fmpz_poly = LazyImport('sage.libs.flint.fmpz_poly_sage', 'Fmpz_poly',  deprecation=36449)



# sage_setup: distribution = sagemath-modules
"""
Root system data for (untwisted) type D affine
"""
#*****************************************************************************
#       Copyright (C) 2008-2009 Daniel Bump
#       Copyright (C) 2008-2009 Justin Walker
#       Copyright (C) 2008-2009 Nicolas M. Thiery <nthiery at users.sf.net>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from .cartan_type import CartanType_standard_untwisted_affine, CartanType_simply_laced


class CartanType(CartanType_standard_untwisted_affine, CartanType_simply_laced):
    def __init__(self, n):
        """
        EXAMPLES::

            sage: ct = CartanType(['D',4,1])
            sage: ct
            ['D', 4, 1]
            sage: ct._repr_(compact = True)
            'D4~'

            sage: ct.is_irreducible()
            True
            sage: ct.is_finite()
            False
            sage: ct.is_affine()
            True
            sage: ct.is_untwisted_affine()
            True
            sage: ct.is_crystallographic()
            True
            sage: ct.is_simply_laced()
            True
            sage: ct.classical()
            ['D', 4]
            sage: ct.dual()
            ['D', 4, 1]

        TESTS::

            sage: TestSuite(ct).run()
        """
        assert n >= 3
        CartanType_standard_untwisted_affine.__init__(self, "D", n)

    def dynkin_diagram(self):
        """
        Return the extended Dynkin diagram for affine type D.

        EXAMPLES::

           sage: d = CartanType(['D', 6, 1]).dynkin_diagram(); d                        # needs sage.graphs
              0 O       O 6
                |       |
                |       |
            O---O---O---O---O
            1   2   3   4   5
            D6~
           sage: d.edges(sort=True)                                                     # needs sage.graphs
           [(0, 2, 1), (1, 2, 1), (2, 0, 1), (2, 1, 1), (2, 3, 1),
            (3, 2, 1), (3, 4, 1), (4, 3, 1), (4, 5, 1), (4, 6, 1), (5, 4, 1), (6, 4, 1)]

           sage: d = CartanType(['D', 4, 1]).dynkin_diagram(); d                        # needs sage.graphs
               O 4
               |
               |
           O---O---O
           1   |2  3
               |
               O 0
           D4~
           sage: d.edges(sort=True)                                                     # needs sage.graphs
           [(0, 2, 1),
            (1, 2, 1),
            (2, 0, 1),
            (2, 1, 1),
            (2, 3, 1),
            (2, 4, 1),
            (3, 2, 1),
            (4, 2, 1)]

           sage: d = CartanType(['D', 3, 1]).dynkin_diagram(); d                        # needs sage.graphs
           0
           O-------+
           |       |
           |       |
           O---O---O
           3   1   2
           D3~
           sage: d.edges(sort=True)                                                     # needs sage.graphs
           [(0, 2, 1), (0, 3, 1), (1, 2, 1), (1, 3, 1),
            (2, 0, 1), (2, 1, 1), (3, 0, 1), (3, 1, 1)]
        """
        from .dynkin_diagram import DynkinDiagram_class
        n = self.n
        if n == 3:
            from . import cartan_type
            res = cartan_type.CartanType(["A",3,1]).relabel({0:0, 1:3, 2:1, 3: 2}).dynkin_diagram()
            res._cartan_type = self
            return res
        g = DynkinDiagram_class(self)
        for i in range(1, n-1):
            g.add_edge(i, i+1)
        g.add_edge(n-2,n)
        g.add_edge(0,2)
        return g

    def _latex_dynkin_diagram(self, label=None, node=None, node_dist=2, dual=False):
        r"""
        Return a latex representation of the Dynkin diagram.

        EXAMPLES::

            sage: print(CartanType(['D',4,1])._latex_dynkin_diagram())
            \draw (0,0.7 cm) -- (2 cm,0);
            \draw (0,-0.7 cm) -- (2 cm,0);
            \draw (2 cm,0) -- (2 cm,0);
            \draw (2 cm,0) -- (4 cm,0.7 cm);
            \draw (2 cm,0) -- (4 cm,-0.7 cm);
            \draw[fill=white] (0 cm, 0.7 cm) circle (.25cm) node[left=3pt]{$0$};
            \draw[fill=white] (0 cm, -0.7 cm) circle (.25cm) node[left=3pt]{$1$};
            \draw[fill=white] (2 cm, 0 cm) circle (.25cm) node[below=4pt]{$2$};
            \draw[fill=white] (4 cm, 0.7 cm) circle (.25cm) node[right=3pt]{$4$};
            \draw[fill=white] (4 cm, -0.7 cm) circle (.25cm) node[right=3pt]{$3$};
            <BLANKLINE>
        """
        if label is None:
            label = lambda i: i
        if node is None:
            node = self._latex_draw_node
        n = self.n
        if n == 3:
            from . import cartan_type
            relabel = {0:label(0), 1:label(3), 2:label(1), 3:label(2)}
            return cartan_type.CartanType(["A",3,1]).relabel(relabel)._latex_dynkin_diagram(node_dist=node_dist)
        rt_most = (n - 2) * node_dist
        center_point = rt_most - node_dist
        ret = "\\draw (0,0.7 cm) -- (%s cm,0);\n" % node_dist
        ret += "\\draw (0,-0.7 cm) -- (%s cm,0);\n" % node_dist
        ret += "\\draw (%s cm,0) -- (%s cm,0);\n" % (node_dist, center_point)
        ret += "\\draw (%s cm,0) -- (%s cm,0.7 cm);\n" % (center_point, rt_most)
        ret += "\\draw (%s cm,0) -- (%s cm,-0.7 cm);\n" % (center_point, rt_most)
        ret += node(0, 0.7, label(0), "left=3pt")
        ret += node(0, -0.7, label(1), "left=3pt")
        for i in range(1, self.n-2):
            ret += node(i*node_dist, 0, label(i+1))
        ret += node(rt_most, 0.7, label(n), "right=3pt")
        ret += node(rt_most, -0.7, label(n-1), "right=3pt")
        return ret

    def ascii_art(self, label=None, node=None):
        """
        Return an ascii art representation of the extended Dynkin diagram.

        TESTS::

            sage: print(CartanType(['D',6,1]).ascii_art(label = lambda x: x+2))
              2 O       O 8
                |       |
                |       |
            O---O---O---O---O
            3   4   5   6   7

            sage: print(CartanType(['D',4,1]).ascii_art(label = lambda x: x+2))
                O 6
                |
                |
            O---O---O
            3   |4  5
                |
                O 2

            sage: print(CartanType(['D',3,1]).ascii_art(label = lambda x: x+2))
            2
            O-------+
            |       |
            |       |
            O---O---O
            5   3   4
        """
        if label is None:
            label = lambda i: i
        if node is None:
            node = self._ascii_art_node
        n = self.n
        if n == 3:
            from . import cartan_type
            return cartan_type.CartanType(["A",3,1]).relabel({0:0, 1:3, 2:1, 3: 2}).ascii_art(label, node)
        if n == 4:
            ret = "    {} {}\n".format(node(label(4)), label(4)) + "    |\n    |\n"
            ret += "{}---{}---{}\n".format(node(label(1)), node(label(2)), node(label(3)))
            ret += "{!s:4}|{!s:3}{!s:4}\n".format(label(1), label(2), label(3))
            ret += "    |\n    {} {}".format(node(label(0)), label(0))
            return ret

        ret = "{!s:>3} {}".format(label(0), node(label(0)))
        ret += (4*(n-4)-1)*" "+"{} {}\n".format(node(label(n)), label(n))
        ret += "    |" + (4*(n-4)-1)*" " + "|\n"
        ret += "    |" + (4*(n-4)-1)*" " + "|\n"
        ret += "---".join(node(label(i)) for i in range(1, n))
        ret += '\n' + "".join("{!s:4}".format(label(i)) for i in range(1,n))
        return ret



# sage_setup: distribution = sagemath-modules
from sage.structure.element cimport AlgebraElement, Element, Vector
from sage.matrix.matrix cimport Matrix

cdef class FiniteDimensionalAlgebraElement(AlgebraElement):
    cdef public Matrix _vector
    cdef Matrix __matrix
    cdef FiniteDimensionalAlgebraElement __inverse

    cpdef dict monomial_coefficients(self, bint copy=*)

cpdef FiniteDimensionalAlgebraElement unpickle_FiniteDimensionalAlgebraElement(A, vec, mat)



# sage_setup: distribution = sagemath-flint
# Functions removed from flint but still needed in Sage. Code adapted from
# earlier versions of flint.

from sage.libs.gmp.mpq cimport *
from sage.libs.flint.fmpz cimport *
from sage.libs.flint.fmpq cimport *
from sage.libs.flint.fmpq_poly cimport *


cdef void fmpq_poly_scalar_mul_mpz(fmpq_poly_t rop, const fmpq_poly_t op, const mpz_t c) noexcept:
    cdef fmpz_t f
    fmpz_init_set_readonly(f, c)
    fmpq_poly_scalar_mul_fmpz(rop, op, f)
    fmpz_clear_readonly(f)

cdef void fmpq_poly_scalar_mul_mpq(fmpq_poly_t rop, const fmpq_poly_t op, const mpq_t c) noexcept:
    cdef fmpq_t f
    fmpq_init_set_readonly(f, c)
    fmpq_poly_scalar_mul_fmpq(rop, op, f)
    fmpq_clear_readonly(f)

cdef void fmpq_poly_set_coeff_mpq(fmpq_poly_t poly, slong n, const mpq_t x) noexcept:
    cdef fmpq_t t
    fmpq_init_set_readonly(t, x)
    fmpq_poly_set_coeff_fmpq(poly, n, t)
    fmpq_clear_readonly(t)

cdef void fmpq_poly_get_coeff_mpq(mpq_t x, const fmpq_poly_t poly, slong n) noexcept:
    cdef fmpq_t t
    fmpq_init(t)
    fmpq_poly_get_coeff_fmpq(t, poly, n)
    fmpq_get_mpq(x, t)
    fmpq_clear(t)

cdef void fmpq_poly_set_mpq(fmpq_poly_t poly, const mpq_t x) noexcept:
    fmpq_poly_fit_length(poly, 1)
    fmpz_set_mpz(fmpq_poly_numref(poly), mpq_numref(x))
    fmpz_set_mpz(fmpq_poly_denref(poly), mpq_denref(x))
    _fmpq_poly_set_length(poly, 1)
    _fmpq_poly_normalise(poly)

cdef void fmpq_poly_set_mpz(fmpq_poly_t poly, const mpz_t x) noexcept:
    fmpq_poly_fit_length(poly, 1)
    fmpz_set_mpz(fmpq_poly_numref(poly), x)
    fmpz_one(fmpq_poly_denref(poly))
    _fmpq_poly_set_length(poly, 1)
    _fmpq_poly_normalise(poly)



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/fq_vec.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    fq_struct * _fq_vec_init(slong len, const fq_ctx_t ctx) noexcept
    void _fq_vec_clear(fq_struct * vec, slong len, const fq_ctx_t ctx) noexcept
    void _fq_vec_randtest(fq_struct * f, flint_rand_t state, slong len, const fq_ctx_t ctx) noexcept
    int _fq_vec_fprint(FILE * file, const fq_struct * vec, slong len, const fq_ctx_t ctx) noexcept
    int _fq_vec_print(const fq_struct * vec, slong len, const fq_ctx_t ctx) noexcept
    void _fq_vec_set(fq_struct * vec1, const fq_struct * vec2, slong len2, const fq_ctx_t ctx) noexcept
    void _fq_vec_swap(fq_struct * vec1, fq_struct * vec2, slong len2, const fq_ctx_t ctx) noexcept
    void _fq_vec_zero(fq_struct * vec, slong len, const fq_ctx_t ctx) noexcept
    void _fq_vec_neg(fq_struct * vec1, const fq_struct * vec2, slong len2, const fq_ctx_t ctx) noexcept
    bint _fq_vec_equal(const fq_struct * vec1, const fq_struct * vec2, slong len, const fq_ctx_t ctx) noexcept
    bint _fq_vec_is_zero(const fq_struct * vec, slong len, const fq_ctx_t ctx) noexcept
    void _fq_vec_add(fq_struct * res, const fq_struct * vec1, const fq_struct * vec2, slong len2, const fq_ctx_t ctx) noexcept
    void _fq_vec_sub(fq_struct * res, const fq_struct * vec1, const fq_struct * vec2, slong len2, const fq_ctx_t ctx) noexcept
    void _fq_vec_scalar_addmul_fq(fq_struct * vec1, const fq_struct * vec2, slong len2, const fq_t c, const fq_ctx_t ctx) noexcept
    void _fq_vec_scalar_submul_fq(fq_struct * vec1, const fq_struct * vec2, slong len2, const fq_t c, const fq_ctx_t ctx) noexcept
    void _fq_vec_dot(fq_t res, const fq_struct * vec1, const fq_struct * vec2, slong len2, const fq_ctx_t ctx) noexcept



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.geometry.polyhedron
"""
Diamond cutting implementation

AUTHORS:

- Jan Poeschko (2012-07-02): initial version
"""
# ****************************************************************************
#       Copyright (C) 2012 Jan Poeschko <jan@poeschko.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.geometry.polyhedron.constructor import Polyhedron
from sage.matrix.constructor import matrix
from sage.modules.free_module_element import vector

from math import sqrt, floor, ceil


def plane_inequality(v) -> list:
    """
    Return the inequality for points on the same side as the origin
    with respect to the plane through ``v`` normal to ``v``.

    EXAMPLES::

        sage: from sage.modules.diamond_cutting import plane_inequality
        sage: ieq = plane_inequality([1, -1]); ieq
        [2, -1, 1]
        sage: ieq[0] + vector(ieq[1:]) * vector([1, -1])
        0
    """
    v = vector(v)
    c = -v * v
    if c < 0:
        c, v = -c, -v
    return [c] + list(v)


def jacobi(M):
    r"""
    Compute the upper-triangular part of the Cholesky/Jacobi
    decomposition of the symmetric matrix ``M``.

    Let `M` be a symmetric `n \times n`-matrix over a field `F`.
    Let `m_{i,j}` denote the `(i,j)`-th entry of `M` for any
    `1 \leq i \leq n` and `1 \leq j \leq n`. Then, the
    upper-triangular part computed by this method is the
    upper-triangular `n \times n`-matrix `Q` whose
    `(i,j)`-th entry `q_{i,j}` satisfies

    .. MATH::

        q_{i,j} =
        \begin{cases}
            \frac{1}{q_{i,i}} \left( m_{i,j} - \sum_{r<i} q_{r,r} q_{r,i} q_{r,j} \right) & i < j, \\
            m_{i,j} - \sum_{r<i} q_{r,r} q_{r,i}^2 & i = j, \\
            0 & i > j,
        \end{cases}

    for all `1 \leq i \leq n` and `1 \leq j \leq n`. (These
    equalities determine the entries of `Q` uniquely by
    recursion.) This matrix `Q` is defined for every invertible
    `n \times n`-matrix `M`. Its definition is taken from (2.3)
    of [FP1985]_.

    .. NOTE::

        This should be a method of matrices.

    EXAMPLES::

        sage: from sage.modules.diamond_cutting import jacobi
        sage: jacobi(identity_matrix(3) * 4)
        [4 0 0]
        [0 4 0]
        [0 0 4]

        sage: def testall(M):
        ....:      Q = jacobi(M)
        ....:      for j in range(3):
        ....:          for i in range(j):
        ....:              if Q[i,j] * Q[i,i] != M[i,j] - sum(Q[r,i] * Q[r,j] * Q[r,r] for r in range(i)):
        ....:                  return False
        ....:      for i in range(3):
        ....:          if Q[i,i] != M[i,i] - sum(Q[r,i] ** 2 * Q[r,r] for r in range(i)):
        ....:              return False
        ....:          for j in range(i):
        ....:              if Q[i,j] != 0:
        ....:                  return False
        ....:      return True

        sage: M = Matrix(QQ, [[8,1,5], [1,6,0], [5,0,3]])
        sage: Q = jacobi(M); Q
        [    8   1/8   5/8]
        [    0  47/8 -5/47]
        [    0     0 -9/47]
        sage: testall(M)
        True

        sage: M = Matrix(QQ, [[3,6,-1,7],[6,9,8,5],[-1,8,2,4],[7,5,4,0]])
        sage: testall(M)
        True
    """
    if not M.is_square():
        raise ValueError("the matrix must be square")
    dim = M.nrows()
    q = [list(row) for row in M]
    for i in range(dim - 1):
        for j in range(i + 1, dim):
            q[j][i] = q[i][j]
            q[i][j] = q[i][j] / q[i][i]
        for k in range(i + 1, dim):
            for l in range(k, dim):
                q[k][l] -= q[k][i] * q[i][l]
    for i in range(1, dim):
        for j in range(i):
            q[i][j] = 0
    return matrix(q)


def diamond_cut(V, GM, C, verbose=False) -> Polyhedron:
    r"""
    Perform diamond cutting on polyhedron ``V`` with basis matrix ``GM``
    and squared radius ``C``.

    INPUT:

    - ``V`` -- polyhedron to cut from

    - ``GM`` -- half of the basis matrix of the lattice

    - ``C`` -- square of the radius to use in cutting algorithm

    - ``verbose`` -- boolean (default: ``False``); whether to print
      debug information

    OUTPUT: a :class:`Polyhedron` instance

    ALGORITHM:

    Use the algorithm in (2.8) of [FP1985]_ to iterate through the nonzero
    vectors ``hv`` of length at most `\sqrt{C}` in the lattice spanned by
    ``GM``. (Actually, the algorithm only constructs one vector from each pair
    ``{hv, -hv}``.) For each such vector ``hv``, intersect ``V`` with the
    half-spaces defined by ``plane_inequality(hv)`` and
    ``plane_inequality(-hv)``.

    EXAMPLES::

        sage: from sage.modules.diamond_cutting import diamond_cut
        sage: V = Polyhedron([[0], [2]])
        sage: GM = matrix([2])
        sage: V = diamond_cut(V, GM, 4)
        sage: V.vertices()
        (A vertex at (2), A vertex at (0))

    TESTS:

    Verify that code works when no cuts are performed::

        sage: from sage.modules.free_module_integer import IntegerLattice
        sage: v = vector(ZZ, [1,1,-1])
        sage: L = IntegerLattice([v])
        sage: C = L.voronoi_cell(radius=0.1)
    """
    if verbose:
        print("Cut\n{}\nwith squared radius {}".format(GM, C))

    dim = GM.dimensions()
    if dim[0] != dim[1]:
        raise ValueError("the matrix must be square")
    dim = dim[0]
    T = [0] * dim
    U = [0] * dim
    x = [0] * dim
    L = [0] * dim

    # calculate the Gram matrix
    q = matrix([[sum(GM[i][k] * GM[j][k] for k in range(dim))
                 for j in range(dim)] for i in range(dim)])
    if verbose:
        print("q:\n{}".format(q.n()))
    # apply Cholesky/Jacobi decomposition
    q = jacobi(q)
    if verbose:
        print("q:\n{}".format(q.n()))

    i = dim - 1
    T[i] = C
    U[i] = 0

    new_dimension = True
    cut_count = 0
    inequalities = []
    while True:
        if verbose:
            print(f"Dimension: {i}")
        if new_dimension:
            Z = sqrt(T[i] / q[i][i])
            if verbose:
                print("Z: {}".format(Z))
            L[i] = int(floor(Z - U[i]))
            if verbose:
                print("L: {}".format(L))
            x[i] = int(ceil(-Z - U[i]) - 1)
            new_dimension = False

        x[i] += 1
        if verbose:
            print(f"x: {x}")
        if x[i] > L[i]:
            i += 1
        elif i > 0:
            T[i - 1] = T[i] - q[i][i] * (x[i] + U[i]) ** 2
            i -= 1
            U[i] = 0
            for j in range(i + 1, dim):
                U[i] += q[i][j] * x[j]
            new_dimension = True
        else:
            if all(elmt == 0 for elmt in x):
                break
            hv = [0] * dim
            for k in range(dim):
                for j in range(dim):
                    hv[k] += x[j] * GM[j][k]
            hv = vector(hv)

            for hv in [hv, -hv]:
                cut_count += 1
                if verbose:
                    print("\n%d) Cut using normal vector %s" % (cut_count, hv))
                inequalities.append(plane_inequality(hv))

    if verbose:
        print("Final cut")
    if inequalities:
        cut = Polyhedron(ieqs=inequalities)
        V = V.intersection(cut)

    if verbose:
        print("End")

    return V


def calculate_voronoi_cell(basis, radius=None, verbose=False) -> Polyhedron:
    """
    Calculate the Voronoi cell of the lattice defined by basis.

    INPUT:

    - ``basis`` -- embedded basis matrix of the lattice

    - ``radius`` -- square of radius of basis vectors to consider

    - ``verbose`` -- whether to print debug information

    OUTPUT: a :class:`Polyhedron` instance

    EXAMPLES::

        sage: from sage.modules.diamond_cutting import calculate_voronoi_cell
        sage: V = calculate_voronoi_cell(matrix([[1, 0], [0, 1]]))
        sage: V.volume()
        1

    TESTS:

    Verify that :issue:`39507` is fixed::

        sage: from sage.modules.free_module_integer import IntegerLattice
        sage: v = vector(ZZ, [1,1,1,-1])
        sage: L = IntegerLattice([v])
        sage: print(v in L)
        True
        sage: print(L.closest_vector(v))
        (1, 1, 1, -1)
        sage: C = L.voronoi_cell()
        sage: C.Hrepresentation()
        (An inequality (-1, -1, -1, 1) x + 2 >= 0,
         An inequality (1, 1, 1, -1) x + 2 >= 0)
        sage: v = vector(ZZ, [1,1,-1])
        sage: L = IntegerLattice([v])
        sage: C = L.voronoi_cell()
        sage: C.Hrepresentation()
        (An inequality (-2, -2, 2) x + 3 >= 0,
         An inequality (2, 2, -2) x + 3 >= 0)
        sage: C.Vrepresentation()
        (A line in the direction (0, 1, 1),
         A line in the direction (1, 0, 1),
         A vertex at (0, 0, -3/2),
         A vertex at (0, 0, 3/2))

    Verify that :issue:`37086` is fixed::

        sage: from sage.modules.free_module_integer import IntegerLattice
        sage: l  = [7, 0, -1, -2, -1, -2, 7, -2, 0, 0, -2,
        ....:       0, 7, -2, 0, -1, -2, -1, 7, 0 , -1, -1, 0, -2, 7]
        sage: M = matrix(5, 5, l)
        sage: C = IntegerLattice(M).voronoi_cell()
        sage: C
        A 5-dimensional polyhedron in QQ^5 defined as the
        convex hull of 720 vertices
    """
    dim = basis.dimensions()
    # LLL-reduce for efficiency.
    basis = basis.LLL()
    if radius is None:
        # Convert the basis matrix to use RDF numbers for efficiency when we
        # calculate the triangular matrix of the QR decomposition.
        from sage.rings.real_double import RDF
        transposed_RDF_matrix = (basis.transpose()).change_ring(RDF)
        R = transposed_RDF_matrix.QR()[1]
        # The length of the vector formed by the diagonal entries of R is an
        # upper bound for twice the covering radius, so it is an upper bound
        # on the length of the lattice vectors that need to be considered for
        # diamond cutting. However, the value of the `radius` keyword is
        # actually a squared length, so there is no square root in the
        # following formula.
        radius = sum(R[i, i]**2 for i in range(dim[0]))
        # We then divide by 4 as we will divide the basis by 2 later on.
        radius = ceil(radius / 4)
    artificial_length = None
    if dim[0] < dim[1]:
        F = basis.base_ring().fraction_field()
        # Introduce "artificial" basis points (representing infinity).
        additional_vectors = (F**dim[1]).subspace(basis).complement().basis()
        additional_vectors = matrix(additional_vectors)
        # LLL-reduce for efficiency.
        additional_vectors = additional_vectors.LLL()

        from sage.rings.real_double import RDF
        # Convert the basis matrix to use RDF numbers for efficiency when we
        # perform the QR decomposition.
        transposed_RDF_matrix = additional_vectors.transpose().change_ring(RDF)
        R = transposed_RDF_matrix.QR()[1]
        # Since R is triangular, its smallest diagonal entry provides a
        # lower bound on the length of the shortest nonzero vector in the
        # lattice spanned by the artificial points. We square it because
        # value of `radius` is a squared length.
        shortest_vector_lower_bound = min(R[i, i]**2
                                          for i in range(dim[1] - dim[0]))
        # We will multiply our artificial points by the following scalar in
        # order to make sure the squared length of the shortest
        # nonzero vector is greater than radius, even after the vectors
        # are divided by 2.
        artificial_length = ceil(2.001 * sqrt(radius / shortest_vector_lower_bound))
        additional_vectors *= artificial_length
        basis = basis.stack(additional_vectors)
        basis = matrix([v for v in basis if v])
        dim = basis.dimensions()
    if dim[0] != dim[1]:
        raise ValueError("invalid matrix")
    basis = basis / 2

    ieqs = []
    for v in basis:
        ieqs.append(plane_inequality(v))
        ieqs.append(plane_inequality(-v))
    Q = Polyhedron(ieqs=ieqs)

    V = diamond_cut(Q, basis, radius, verbose=verbose)

    if artificial_length is not None:
        # Remove inequalities introduced by artificial basis points.
        H = V.Hrepresentation()
        H = [v for v in H if all(not V._is_zero(v.A() * w / 2 - v.b()) and
                                 not V._is_zero(v.A() * (-w) / 2 - v.b())
                                 for w in additional_vectors)]
        V = Polyhedron(ieqs=H)

    return V
